from collections import defaultdict
from copy import copy
from typing import Dict, List, Tuple, Union

from SQcircuit import Circuit, Element
from SQcircuit.settings import get_optim_mode
import torch

SQValType = Union[float, torch.Tensor]


def construct_perturbed_elements(
        circuit: Circuit,
        new_values: torch.Tensor,
) -> Dict[Tuple[int, int], List[Element]]:
    """
    Construct a dictionary of elements based on ``circuit``, where the elements
    present in `circuit.parameters`` have their values updated by those
    provided in ``new_values``.

    Parameters
    ----------
        circuit:
            A circuit to use as the initial elements.
        new_values:
            The values for the perturbed elements to take. This is a tensor of
            length ``circuit.parameters``, and the values are assigned to the 
            elements according to the order given by
            ``circuit.parameters_elems``.
    
    Returns
    ----------
        A dictionary of elements, with perturbed values.
    """
    # Construct replacement loops with updated values, where necessary
    replacement_loops = {}
    for loop in circuit.loops:
        if loop in circuit.parameters:
            new_loop = copy(loop)
            new_loop.internal_value = new_values[
                circuit.parameters_elems.index(loop)
            ]
        else:
            replacement_loops[loop] = loop

    new_elements = defaultdict(list)
    for edge in circuit.elements:
        for el in circuit.elements[edge]:
            new_el = copy(el)

            # Update value, if necessary
            if el in circuit.parameters_elems:
                new_el.internal_value = new_values[
                    circuit.parameters_elems.index(el)
                ]

            # Update loop values, if necessary
            if hasattr(el, 'loops'):
                new_loops = []
                for loop in el.loops:
                    new_loops.append(replacement_loops[loop])
                new_el.loops = new_loops

            new_elements[edge].append(new_el)

    return new_elements


def detach_if_optim(value: SQValType) -> SQValType:
    """Detach the value if is in torch. Otherwise, return the value itself."""

    if get_optim_mode():
        return value.detach()

    return value


def hinge_loss(val: SQValType, cutoff: float, slope: float) -> SQValType:
    """
    Compute a linear hinge loss.
    
    Parameters
    ----------
        val:
            The input value.
        cutoff:
            The cutoff above which to apply hinge loss.
        slope:
            The slope of the loss above ``cutoff``.
    
    Returns
    ----------
        Zero if ``val < cutoff``, otherwise ``slope * (val - cutoff)``.      
    """
    if val < cutoff:
        return 0.0 * val
    else:
        return slope * (val - cutoff)


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def zero() -> SQValType:
    """
    Return the value of 0 in the appropriate datatype for the current
    optimization mode.

    Returns
    ----------
        The float 0.0 if not get_optim_mode(), otherwise the tensor value of 0.
    """
    if get_optim_mode():
        return torch.tensor(0.0)

    return 0.0



"""Contains code for defining loss functions used in circuit optimization."""
from typing import Callable, Dict, List, Optional, Tuple

from SQcircuit import Circuit
from SQcircuit.settings import get_optim_mode
import torch

from .functions import (
    charge_sensitivity,
    flux_sensitivity,
    element_sensitivity,
    first_resonant_frequency,
    reset_charge_modes,
    zero,
    decoherence_time,
    total_dec_time,
    fastest_gate_speed,
    number_of_gates
)
from .utils import (
    detach_if_optim,
    dotdict,
    hinge_loss,
    SQValType
)

# Added on to a loss when it is close to zero, so that loss values are never
# exactly zero (to allow logarithms, etc.)
EPSILON = 1e-13

def t1_loss(
    circuit: Circuit,
    dec_type: str = 'total',
) -> Tuple[SQValType, SQValType]:
    """Penalizes short T1 time.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        dec_type:
            A decoherence channel to consider, or ``'total'`` to use all
            available.

    Returns
    ----------
        loss:
            ``-T1``.
        T1:
            The T1 time of ``circuit``.
    """
    t1 = decoherence_time(
        circuit=circuit,
        t_type='t1',
        dec_type=dec_type
    )

    return -t1 + EPSILON, t1


def t_phi_loss(
        circuit: Circuit,
        dec_type='total'
) -> Tuple[SQValType, SQValType]:
    """Penalizes short dephasing time.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        dec_type:
            A dephasing channel to consider, or ``'total'`` to use all
            available.

    Returns
    ----------
        loss:
            ``-T_phi``.
        T_phi:
            The dephasing time of ``circuit``.
    """
    t_phi = decoherence_time(
        circuit=circuit,
        t_type='t_phi',
        dec_type=dec_type
    )

    return -t_phi + EPSILON, t_phi


def t2_loss(circuit: Circuit) -> Tuple[SQValType, SQValType]:
    """Penalizes short T2 time, computed via all available decay channels.
    See ``functions.total_dec_time`` for a description of the decay channels
    considered.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.s

    Returns
    ----------
        loss:
            ``-T2``.
        T2:
            The T2 time of ``circuit``.
    """
    t2 = total_dec_time(circuit)

    return -t2 + EPSILON, t2


def element_sensitivity_loss(
    circuit: Circuit,
    n_samples=100,
    error=0.01,
) -> Tuple[SQValType, SQValType]:
    """Computes the sensitivity of the single-qubit gate number of the qubit
    with respect to variation in the element values. See
    ``functions.element_sensitivity`` for more details on how this is computed.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        n_samples:
            The number of randomly sampled circuits to calculate. More samples
            provide more reproducible results.
        fabrication_error:
            The percentage fabrication error to simulate.

    Returns
    ----------
        loss:
            The sensitivity ``sens``.
        sensitivity:
            The sensitivity of the gate number to element variation.
    """
    sens = element_sensitivity(circuit, number_of_gates, n_samples, error)

    return sens + EPSILON, sens


def gate_speed_loss(circuit: Circuit) -> Tuple[SQValType, SQValType]:
    """Penalizes a slow single-qubit gate speed for the circuit. See
    ``functions.fastest_gate_speed`` for more details on how this is computed.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.

    Returns
    ----------
        loss:
            1 / ``gate_speed``.
        gate_speed:
            The gate speed of ``circuit``.
    
    """
    gate_speed = fastest_gate_speed(circuit)

    return (1 / gate_speed) + EPSILON, gate_speed


def anharmonicity_loss(
    circuit: Circuit
) -> Tuple[SQValType, SQValType]:
    """Computes a loss to penalize low anharmonicitiy. The absolute anharmonicity is
        ``A = omega_{21} - omega_{10}``
    and the relative anharmonicity is
        ``Ar = (omega_{21} - omega_{10})/omega_{10}``.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.

    Returns
    ----------
        loss:
            1 / Ar, the reciprocal of the relative anharmonicity.
        anharmonicity:
            A, the absolute anharmonicity.
    """

    omega_10 = circuit.efreqs[1] - circuit.efreqs[0]
    omega_21 = circuit.efreqs[2] - circuit.efreqs[1]

    A = omega_21 - omega_10

    return (omega_10 / A) + EPSILON, A


def frequency_loss(
        circuit: Circuit,
        freq_threshold: float = 100.0,
) -> Tuple[SQValType, SQValType]:
    """Loss function which applies a quadratic hinge loss to qubits whose
    frequency exceeds ``freq_threshold``. When ``f_{10} < freq_threshold``, the
    loss is zero, and otherwise it is ``(f_{10} - freq_threshold)**2``.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        freq_thresholdD:
            The frequency threshold (in the frequency unit of SQcircuit)
            above which to apply the quadratic hinge loss. Default 100 GHz.

    Returns
    ----------
        loss:
            The hinge loss applied to ``f``.
        f:
            The qubit frequency of ``circuit``.
     """
    freq = first_resonant_frequency(circuit)
    if freq > freq_threshold:
        loss = (freq - freq_threshold)**2
    else:
        loss = zero()
    return loss + EPSILON, freq


def flux_sensitivity_loss(
    circuit: Circuit,
    a=0.1,
    b=1,
) -> Tuple[SQValType, SQValType]:
    """Applies a hinge loss to flux sensitivity of circuit. The flux sensitivity
    is the normalized variation in qubit frequency for small perturbations of
    the external flux about the operating point. See ``functions.flux_sensitivity``
    for more details.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        a:
            The cutoff above which to apply the hinge loss.
        b:
            The slope for the hinge loss.

    Returns
    ----------
        loss:
            The hinge loss applied to ``S``. 
        S:
            The flux sensitivity of ``circuit``.
    """

    sens = flux_sensitivity(circuit)

    return hinge_loss(sens, a, b) + EPSILON, sens


def charge_sensitivity_loss(
    circuit: Circuit,
    a=0.02,
    b=1,
) -> Tuple[SQValType, SQValType]:
    """Applies a linear hinge loss to charge sensitivity of circuit. See
    ``functions.charge_sensitivity`` for details on how the charge sensitivity
    is calculated.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.
        a:
            The cutoff above which to apply the hinge loss.
        b:
            The slope for the hinge loss.

    Returns
    ----------
        loss:
            The hinge loss applied to ``S``. 
        S:
            The charge sensitivity of ``circuit``.
    """

    sens = charge_sensitivity(circuit)

    reset_charge_modes(circuit)
    return hinge_loss(sens, a, b) + EPSILON, sens


def number_of_gates_loss(
    circuit: Circuit,
) -> Tuple[SQValType, SQValType]:
    """Applies a loss ``1/N`` to the number of single-qubit gates ``N`` for the
    circuit. See ``functions.number_of_gates`` for details on how the number
    of gates is calculated.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.

    Returns
    ----------
        loss:
            The loss ``1/N``.
        N:
            The number of single-qubit gates.
    """

    N = number_of_gates(circuit)

    loss = 1 / N

    return loss, N


###############################################################################
# Incorporating all losses into one loss function
###############################################################################

ALL_METRICS = dotdict({
    ###########################################################################
    'frequency': frequency_loss,
    'flux_sensitivity': flux_sensitivity_loss,
    'charge_sensitivity': charge_sensitivity_loss,
    'number_of_gates': number_of_gates_loss,
    ###########################################################################
    'anharmonicity': anharmonicity_loss,
    'element_sensitivity': element_sensitivity_loss,
    'gate_speed': gate_speed_loss,
    't2': t2_loss,
    't1': t1_loss,
    't1_capacitive': lambda cr: t1_loss(cr, dec_type='capacitive'),
    't1_inductive': lambda cr: t1_loss(cr, dec_type='inductive'),
    't1_quasiparticle': lambda cr: t1_loss(cr, dec_type='quasiparticle'),
    't_phi': t_phi_loss,
    't_phi_charge': lambda cr: t_phi_loss(cr, dec_type='charge'),
    't_phi_cc': lambda cr: t_phi_loss(cr, dec_type='cc'),
    't_phi_flux': lambda cr: t_phi_loss(cr, dec_type='flux')
})


def get_all_metrics() -> List[str]:
    """
    Provides a list of all available metrics which can be calculated for a
    circuit or used to construct a loss function. Each function can be accessed
    via ``ALL_METRICS[key]``, where ``key`` is any string in the returned
    list.

    Returns
    ----------
        List of names of available metrics.
    """
    return list(ALL_METRICS.keys())


def describe_metric(name: str) -> None:
    """
    Provides a description of a given metric function (if available).

    Parameters
    ----------
        Name of metric.
    """
    descrip = ALL_METRICS[name].__doc__

    if descrip is None:
        print('No description available.')
    else:
        print(descrip)


def add_to_metrics(
    name: str,
    function: Callable[[Circuit], Tuple[SQValType, SQValType]]
) -> None:
    """Add a function to the available metrics to construct a loss function.

    Parameters
    ----------
        name:
            Name of new metric.
        function:
            Function computing the metric and loss.
    """
    ALL_METRICS[name] = function


LossFunctionRetType = Tuple[SQValType, Dict[str, SQValType], Dict[str, SQValType]]
LossFunctionType = Callable[
    [Circuit],
    LossFunctionRetType
]


def build_loss_function(
    use_losses: Dict[str, float],
    use_metrics: Optional[List[str]] = None,
    master_use_grad: bool = True
) -> LossFunctionType:
    """Build a loss function based on the metrics provided in ``use_losses``.
    The constructed loss function will also calculate metrics in
    ``use_metrics``.
    
    The loss function takes in a circuit and returns
        - Total loss: the sum of metrics in ``use_losses``.
        - Loss values: A dictionary of ``metric: value`` for metrics in
        ``use_losses``.
        - Metric values: A dictionary of ``metric: value`` for metrics in
        ``use_metrics``.

    Parameters
    ----------
        use_losses:
            A dictionary of ``metric: weight`` pairs, where ``metric`` is the 
            name of a metric returned by ``get_all_metrics()`` and ``weight`` 
            is the weight in the total loss function.
        use_metrics:
            A list of metrics to calculate, where the available metrics are 
            provided in the list returned by ``get_all_metrics()``.
        master_use_grad:
            Whether to enable gradient when calculating the total loss.

    Returns
    ----------
        A loss function which computes the total loss using ``use_losses`` as
        well as the metrics in ``use_metrics``.
    """
    if use_metrics is None:
        use_metrics = []

    if not isinstance(use_losses, dict):
        raise ValueError('You must pass in dictionary of metric_name: weight'
                         ' for `use_losses`.')
    for metric_name in use_losses.keys():
        if metric_name not in get_all_metrics():
            raise ValueError(f'The metric \'{metric_name}\' is not available.')
    if not isinstance(use_metrics, list):
        raise ValueError('You must pass in list of metrics for `use_metrics`.')
    for metric_name in use_metrics:
        if metric_name not in get_all_metrics():
            raise ValueError(f'The metric \'{metric_name}\' is not available.')

    return lambda circuit: calculate_loss_metrics(
        circuit,
        use_losses,
        use_metrics,
        master_use_grad=master_use_grad
    )


def calculate_loss_metrics(
    circuit: Circuit,
    use_losses: Dict[str, float],
    use_metrics: List[str],
    master_use_grad: bool = True,
) -> LossFunctionRetType:
    """Calculates an overall loss function for ``circuit``, based on metrics 
    provided in ``use_losses``. Additionally calculates other metrics provided
    in ``use_metrics``, but does not include them in the overall loss.

    Parameters
    ----------
        circuit:
            The circuit to calculate the loss/metrics for.
        use_losses:
            A dictionary of ``metric: weight`` pairs, where ``metric`` is the 
            name of a metric returned by ``get_all_metrics()`` and ``weight`` 
            is the weight in the overall loss function.
        use_metrics:
            A list of metrics to calculate, where the available metrics are 
            provided in the list returned by ``get_all_metrics()``.
        master_use_grad:
            Whether to enable gradient when calculating metrics to use in
            the overall loss.

    Returns
    ----------
        total_loss:
            The total loss function, which is ``sum(weight*metric)`` over 
            the metrics in ``use_losses``.
        loss_values
            A dictionary of ``metric: value`` for each metric in
            ``use_losses``.
        metric_values
            A dictionary of ``metric: value`` for each metric in
            ``use_metrics``.
    """

    if get_optim_mode():
        loss = torch.zeros((), requires_grad=master_use_grad)
    else:
        loss = 0.0

    loss_values: Dict[str, SQValType] = {}
    metrics: Dict[str, SQValType] = {}

    try:
        for key in use_losses:
            with torch.set_grad_enabled(master_use_grad):
                specific_loss, specific_metric = ALL_METRICS[key](circuit)
                loss = loss + use_losses[key] * specific_loss
            with torch.no_grad():
                loss_values[key + '_loss'] = detach_if_optim(specific_loss)
                metrics[key] = detach_if_optim(specific_metric)
        for key in use_metrics:
            if key not in use_losses:
                with torch.no_grad():
                    specific_loss, specific_metric = ALL_METRICS[key](circuit)

                    loss_values[key + '_loss'] = detach_if_optim(specific_loss)
                    metrics[key] = detach_if_optim(specific_metric)
    except KeyError as e:
        raise ValueError(f'The metric {e} is not available!') from None

    loss_values['total_loss'] = detach_if_optim(loss)

    return loss, loss_values, metrics



"""This module provides functions to input and output of the scripts.

################################################################################
YAML Configuration File Structure

- name [str]: A unique identifier for the optimization setup.
- K [int]: Total truncation number for circuit quantization.
- epochs [int]: Total number of optimization iteration.
- num_eigenvalues [int]: Total number of eigenvalues for diagonalization.
- capacitor_range [list[float]] Range of capacitor values in Farads, specified
  as [min, max].
- inductor_range [list[float]]: Range of inductor values in Henries, specified
  as [min, max].
- junction_range [list[float]]: Range of junction frequencies in Hertz,
  specified as [min, max].
- use_losses [dict[str, float]]: Specifies which loss components use with float
  values for the weight in the total loss function:
- use_metrics [list[str]]: Metrics to use in simulations, e.g., ["T1"] for
  coherence time.

Example:
    name: "JL"
    K: 120
    epochs: 100
    capacitor_range: [1e-15, 12e-12]
    inductor_range: [1e-15, 5e-6]
    junction_range: [1e9, 100e9]
    num_eigenvalues: 10
    use_losses:
      frequency: 1.0
      anharmonicity: 1.0
      flux_sensitivity: 1.0
      charge_sensitivity: 1.0
      T1: 0.0
    use_metrics: ["T1"]
    optim_type: SGD
    circuit_code: JL

################################################################################
The working tree for the outputs of scripts will look like:

 main_folder/
 ├── yaml_file.yaml
 │
 └── {name}/
     │
     ├── records/
     │   │
     │   ├── loss_record_{circuit_code}_{name}_{optim_type}_{seed}.pickle
     │   ├── metrics_record_{circuit_code}_{name}_{optim_type}_{seed}.pickle
     │   └── circuit_record_{circuit_code}_{name}_{optim_type}_{seed}.pickle
     │
     ├── plots/
     │   │
     │   ├── loss_{circuit_code}_{name}_{optim_type}_n_{num_runs}.png
     │   └── metrics_{circuit_code}_{name}_{optim_type}_n_{num_runs}.png
     │
     └── summaries/
         │
         └── circuit_summary_{circuit_code}_{name}_{optim_type}_{seed}.txt
"""

from collections import defaultdict
import os
from typing import Dict, List, Tuple

import yaml

################################################################################
# General Settings.
################################################################################

# Keys that must be included in YAML file.
YAML_KEYS = [
    'name',
    'K',
    "epochs",
    "num_eigenvalues",
    "capacitor_range",
    "inductor_range",
    "junction_range",
    "use_losses",
    "use_metrics",
]

# unit keys for metrics.
UNITS = {
    'frequency': '[GHz]',
    'gate_speed': '[GHz]',
    't2': '[s]',
    't1': '[s]',
    't1_capacitive': '[s]',
    't1_inductive': '[s]',
    't1_quasiparticle': '[s]',
    't_phi': '[s]',
    't_phi_charge': '[s]',
    't_phi_cc': '[s]',
    't_phi_flux': '[s]',
}
UNITS = defaultdict(lambda: "", UNITS)

################################################################################
# Read Functionalities.
################################################################################


def get_metrics_dict(config: dict) -> Tuple[List[str], List[str]]:
    """Return a list of the metrics that were not used in optimization and the
    metrics that were used in optimization.

    Parameters
    ----------
        config:
            A dictionary containing the parameters of the YAML file.
    """

    metrics_in_optim = list(config['use_losses'].keys())
    metrics_not_in_optim = []
    all_metrics = metrics_in_optim + config['use_metrics']

    for metric in all_metrics:
        if metric not in metrics_in_optim:
            metrics_not_in_optim.append(metric)

    return metrics_in_optim, metrics_not_in_optim


def get_units() -> Dict[str, str]:
    """Return a dictionary containing the units of the metrics"""

    return UNITS


def load_yaml_file(path_to_yaml: str) -> dict:
    """Load yaml file and assert that yaml file has the expected keys.

    Parameters
    ----------
        path_to_yaml:
            A string containing the path to the yaml file.
     """

    with open(path_to_yaml, 'r') as f:
        parameters = yaml.safe_load(f.read())

    for key in YAML_KEYS:
        assert key in parameters.keys(), (
            f'Yaml file must include \'{key}\' key.'
        )

    return parameters


def add_command_line_keys(
    parameters: dict,
    arguments: dict,
    keys: List[str],
    optional_keys: List[str] = None,
) -> dict:
    """Add command line argument keys to the parameters which are loaded from
    YAML file.

    Parameters
    ----------
        parameters:
            A dictionary containing the parameters of the YAML file.
        arguments:
            A dictionary containing the arguments of the command line.
        keys:
            A list of string keys that must be either specified in the YAML
            file or command line options.
        optional_keys:
            A list of string keys that are optional and can be not specified in
            both the YAML file and the command line options.
    """

    if optional_keys is None:
        optional_keys = []

    for key in keys + optional_keys:
        if arguments['--' + key] is not None or key in optional_keys:
            parameters[key] = arguments['--' + key]

        assert key in parameters.keys(), (
            f'\'{key}\' key must be either passed '
            f'in the command line or yaml file'
        )

    return parameters

################################################################################
# Directories
################################################################################


class Directory:
    """Directory class to address the files in the main directories.

    Parameters
    ----------
        parameters:
            A dictionary containing the parameters of YAML file/command line.
        yaml_filepath:
            The location of the YAML file.
    """

    def __init__(self, parameters: dict, yaml_filepath: str) -> None:

        self.parameters = parameters
        self.loc = os.path.dirname(os.path.abspath(yaml_filepath))

    def get_experiment_dir(self) -> str:
        """Return the directory for the experiment folder. The experiment folder
        has the following format:
            ``{name}/``
        """

        experiment_dir = os.path.join(
            self.loc,
            self.parameters['name']
        )

        # create the folder if if does not exist.
        os.makedirs(experiment_dir, exist_ok=True)

        return experiment_dir

    def get_records_dir(self) -> str:
        """Return the directory for the records folder inside the experiment
        folder.
        """

        records_dir = os.path.join(
            self.get_experiment_dir(),
            'records',
        )

        # create the folder if it's not excited.
        os.makedirs(records_dir, exist_ok=True)

        return records_dir

    def get_plots_dir(self) -> str:
        """Return the directory for the plot folder inside the experiment
        folder.
        """

        plots_dir = os.path.join(
            self.get_experiment_dir(),
            'plots'
        )

        # create the folder if it does not exist
        os.makedirs(plots_dir, exist_ok=True)

        return plots_dir

    def get_summaries_dir(self) -> str:
        """Return the directory for the summaries folder inside the experiment
        folder.
        """

        summaries_dir = os.path.join(
            self.get_experiment_dir(),
            'summaries'
        )

        # create the folder if it does not exist
        os.makedirs(summaries_dir, exist_ok=True)

        return summaries_dir

    def get_record_file_dir(
        self,
        record_type: str,
        circuit_code: str,
        idx: int,
    ) -> str:
        """Return the directory for the saved circuit record as pickled file
        in the records_directory.

        Parameters
        ----------
            record_type :
                A string indicating the type of record. It can be either 'loss',
                `metrics` or circuit.
            circuit_code:
                A string specifying circuit_code of the circuit.
            idx:
                An Integer specifying the index of the circuit in the records.
        """

        record_name = (
            f'{record_type}_record_{circuit_code}_{self.parameters["name"]}'
            + f'_{self.parameters["optim_type"]}_{idx}.pickle'
        )
        record_name = ''.join([str(word) for word in record_name])

        return os.path.join(
            self.get_records_dir(),
            record_name,
        )

    def get_summary_file_dir(self, circuit_code: str, idx: int) -> str:
        """Return the text file directory to save summary files inside summary
        folder.

        Parameters
        ----------
            circuit_code:
                A string specifying circuit_code of the circuit.
            idx:
                An integer specifying the index of the circuit in the records.
        """
        summary_name = (
            f'circuit_summary_{circuit_code}_{self.parameters["name"]}'
            + f'_{self.parameters["optim_type"]}_{idx}.txt'
        )

        return os.path.join(
            self.get_summaries_dir(),
            summary_name,
        )



from typing import Dict, Optional

from SQcircuit import Circuit, CircuitComponent
from torch.optim import LBFGS
from torch import Tensor

from .optim import OptimizationRecord, run_optimization
from .utils import LossFunctionType

# We set `max_iter = 1` and iterate manually to control print-out.
# This incurs 1 extra function evaluation per step vs. using  the `max_iter`
# in LBFGS itself.
DEFAULT_OPTIM_KWARGS = {
    'lr': 10,
    'history_size': None,
    'max_iter': 1,
    'line_search_fn': 'strong_wolfe',
}

def run_BFGS(
    circuit: Circuit,
    loss_metric_function: LossFunctionType,
    max_iter: int,
    total_trunc_num: int,
    bounds: Dict[CircuitComponent, Tensor],
    optimizer_kwargs: Optional[Dict] = None,
    num_eigenvalues: int = 10,
    save_loc: Optional[str] = None,
    identifier: Optional[str] = None,
    save_intermediate_circuits: bool = False,
) -> OptimizationRecord:
    """Runs BFGS for a maximum of ``max_iter`` beginning with ``circuit`` using
    ``loss_metric_function``.

    Parameters
    ----------
        circuit:
            A circuit to optimize.
        loss_metric_function:
            Loss function to optimize.
        max_iter:
            Maximum number of iterations.
        total_trunc_num:
            Maximum total truncation number to allocate.
        bounds:
            Dictionary giving bounds for each element type.
        num_eigenvalues:
            Number of eigenvalues to calculate when diagonalizing.
        save_loc:
            Folder to save results in, or None.
        identifier:
            String identifying this run (e.g. seed, name, circuit code, …)
            to use when saving.
        save_intermediate_circuits:
            Whether to save the circuit at each iteration.  

    Returns
    ----------
        An ``OptimizationRecord`` containing the optimized circuit, the final
        loss, and a record of the component loss and metric values at each
        epoch.
    """

    # Set up optimizer keyword arguments
    if optimizer_kwargs is None:
        optimizer_kwargs = {}

    optimizer_kwargs = DEFAULT_OPTIM_KWARGS | optimizer_kwargs

    if optimizer_kwargs['history_size'] is None:
        optimizer_kwargs['history_size'] = max_iter


    # Run it!
    return run_optimization(
        circuit = circuit,
        loss_metric_function = loss_metric_function,
        max_iter = max_iter,
        total_trunc_num = total_trunc_num,
        bounds = bounds,
        optimizer = LBFGS,
        uses_closure=True,
        optimizer_kwargs = optimizer_kwargs,
        num_eigenvalues = num_eigenvalues,
        save_loc = save_loc,
        identifier = identifier,
        save_intermediate_circuits = save_intermediate_circuits
    )



"""Test Module for SGD algorithm. CURRENTLY DEPRECATED"""

import pytest
import torch

import SQcircuit as sq

from SQcircuit import Circuit

from qubit_discovery import run_SGD
from qubit_discovery.losses.loss import calculate_loss_metrics
from qubit_discovery.tests.conftest import (
    get_fluxonium,
    get_bounds,
)

@pytest.mark.skip('Currently deprecated')
def test_sgd_run() -> None:
    """Test one step of SGD algorithm."""

    target_params = torch.tensor(
        [1.9388e-14, 1.6453e-07, 6.3300e+09],
        dtype=torch.float64
    )

    def my_loss_function(cr: Circuit):
        return calculate_loss_metrics(
            cr,
            use_losses={
                "frequency": 1.0,
                "anharmonicity": 1.0,
                "flux_sensitivity": 1.0,
                "charge_sensitivity": 1.0,
                "T1": 1.0,
            },
            use_metrics=[],
        )

    sq.set_engine('PyTorch')

    total_trunc_num = 120

    circuit = get_fluxonium()
    circuit.set_trunc_nums([total_trunc_num])

    run_SGD(
        circuit=circuit,
        circuit_code="JL",
        loss_metric_function=my_loss_function,
        name="SGD_test",
        num_eigenvalues=10,
        baseline_trunc_nums=[total_trunc_num],
        total_trunc_num=total_trunc_num,
        num_epochs=1,
        bounds=get_bounds(),
        save_intermediate_circuits=False,
    )

    assert torch.isclose(torch.stack(circuit.parameters), target_params).all()



from SQcircuit.tests.test_grad import function_grad_test
from SQcircuit.tests.conftest import (
    create_fluxonium_numpy,
    create_fluxonium_torch_flux,
    create_flux_transmon_numpy,
    create_flux_transmon_torch
)

from qubit_discovery.losses.functions import (
    flux_sensitivity
)


def test_flux_sensitivity_grad() -> None:
    print('Testing fluxonium')
    flux_points = [1e-2, 0.25, 0.5 - 1e-2, 0.5 + 1e-2, 0.75]
    trunc_num = 120

    for phi_ext in flux_points:
        circuit_numpy = create_fluxonium_numpy(trunc_num, phi_ext)
        circuit_torch = create_fluxonium_torch_flux(trunc_num, phi_ext)

        function_grad_test(
            circuit_numpy,
            flux_sensitivity,
            circuit_torch,
            flux_sensitivity,
            num_eigenvalues=50,
            delta=1e-6
        )

    print('Testing transmon')
    trunc_num = 50

    for phi_ext in flux_points:
        circuit_numpy = create_flux_transmon_numpy(trunc_num, phi_ext)
        circuit_torch = create_flux_transmon_torch(trunc_num, phi_ext)

        function_grad_test(
            circuit_numpy,
            flux_sensitivity,
            circuit_torch,
            flux_sensitivity,
            num_eigenvalues=50,
            delta=1e-6
        )



"""Module contains general configuration and functions for
qubit_discovery tests.
"""

from typing import Dict

import torch
import numpy as np

from torch import Tensor

import SQcircuit as sq
from SQcircuit.settings import get_optim_mode


def get_fluxonium() -> sq.Circuit:
    """Returns a Fluxonium qubit for test purposes."""

    loop = sq.Loop(0.5)
    cap = sq.Capacitor(
        1.0, requires_grad=get_optim_mode()
    )
    ind = sq.Inductor(
        1.0, loops=[loop], requires_grad=get_optim_mode()
    )
    junc = sq.Junction(
        1.0, loops=[loop], requires_grad=get_optim_mode()
    )

    circuit = sq.Circuit(
        elements={
            (0, 1): [cap, ind, junc]
        }
    )

    return circuit

def get_fluxonium_random() -> sq.Circuit:
    """Returns a Fluxonium qubit for test purposes."""

    loop = sq.Loop(0.5)
    cap = sq.Capacitor(
        np.random.normal(1.0, 0.1), requires_grad=get_optim_mode()
    )
    ind = sq.Inductor(
        np.random.normal(1.0, 0.1), loops=[loop], requires_grad=get_optim_mode()
    )
    junc = sq.Junction(
        np.random.normal(1.0, 0.1), loops=[loop], requires_grad=get_optim_mode()
    )

    circuit = sq.Circuit(
        elements={
            (0, 1): [cap, ind, junc]
        }
    )

    return circuit

def get_cpb() -> sq.Circuit:
    """Returns a Transmon qubit for test purposes."""

    loop = sq.Loop(0.5)
    cap = sq.Capacitor(
        1.2, 'GHz', requires_grad=get_optim_mode()
    )
    junc = sq.Junction(
        0.2, 'GHz', loops=[loop], requires_grad=get_optim_mode()
    )

    circuit = sq.Circuit(
        elements={
            (0, 1): [cap, junc]
        }
    )

    return circuit


def get_bounds() -> Dict[sq.Element, Tensor]:
    """Returns element bounds for test purposes."""

    bounds = {
        sq.Capacitor: torch.tensor([1e-15, 12e-12]),
        sq.Inductor: torch.tensor([1e-15, 5e-6]),
        sq.Junction: torch.tensor([1e9, 100e9]),
        sq.Loop: torch.tensor([0, 1])
    }

    return bounds


def are_loss_dicts_close(dict1, dict2, rel: float = 1e-2) -> bool:
    """Check if two loss dictionaries are close in values
     within a given tolerance."""

    if dict1.keys() != dict2.keys():
        return False

    for key in dict1:
        for i in range(len(dict1[key])):
            if not np.isclose(dict1[key][i], dict2[key][i], rtol=rel):
                return False

    return True



"""Test Module for loss helper functions."""

import numpy as np
import torch

import SQcircuit as sq
from SQcircuit.settings import set_optim_mode

from qubit_discovery.losses import (
    element_sensitivity,
    fastest_gate_speed,
    flux_sensitivity,
    number_of_gates
)
from qubit_discovery.tests.conftest import (
    get_fluxonium,
)


def test_fastest_gate_speed() -> None:
    """Test fastest_gate_speed function."""

    # Test settings
    target_omega = torch.tensor(
        [0.55255205],
        dtype=torch.float64
    )
    total_trunc_num = 120
    n_eigs = 10

    def get_circuit_based_op_optim(optimization_mode: bool) -> sq.Circuit:

        set_optim_mode(optimization_mode)

        cr = get_fluxonium()
        cr.set_trunc_nums([total_trunc_num])
        cr.diag(n_eigs)

        return cr

    for optim_mode in [True, False]:

        circuit = get_circuit_based_op_optim(optimization_mode=optim_mode)
        omega = fastest_gate_speed(circuit)

        if optim_mode:
            torch.isclose(omega, target_omega, rtol=1e-2)
        else:
            np.isclose(omega, target_omega.item(), rtol=1e-2)

def test_flux_sensitivity() -> None:
    target_sensitivity = torch.tensor(0.001179668, dtype=torch.float64)

    set_optim_mode(True)
    cr = get_fluxonium()
    cr.set_trunc_nums([120])
    cr.diag(20)
    sens = flux_sensitivity(cr)
    assert torch.isclose(sens, target_sensitivity)

def test_element_sensitivity() -> None:
    target_sensitivity = torch.tensor(0.0178, dtype=torch.float32)

    set_optim_mode(True)
    cr = get_fluxonium()
    cr.set_trunc_nums([120])
    cr.diag(20)
    sens = element_sensitivity(cr, number_of_gates)
    print(sens.dtype)
    assert torch.isclose(sens, target_sensitivity, rtol=0.5)



"""
Plot optimization results.

Usage:
  plot_records.py <yaml_file> [--num_runs=<num_runs>] [--circuit_code=<circuit_code>] 
                  [--optim_type=<optim_type>] [--num_best=<num_best>]
  plot_records.py -h | --help
  plot_records.py --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.

Options:
  -h, --help    Show this screen.
  --version     Show version.

  -n, --num_runs=<num_runs>                 Number of runs to consider
  -c, --circuit_code=<circuit_code>         Circuit code
  -o, --optim_type=<optim_type>             Optimization method
  -b, --num_best=<num_best>                 <num_best> number of best [Optional]
"""

from collections import defaultdict
import os
from typing import Dict, List

from docopt import docopt
from matplotlib import pyplot as plt

from utils import load_record
from inout import (
    add_command_line_keys,
    Directory,
    get_units,
    load_yaml_file
)

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
RECORDS_REQUIRED_KEYS = [
    "num_runs",
    "circuit_code",
    "optim_type",
]
# optional keys.
RECORDS_OPTIONAL_KEYS = ["num_best"]

PLOT_SCHEME = defaultdict(lambda: '#81221B')
PLOT_SCHEME['JL'] = 'darkorange'

# metric keys for plotting.
METRIC_NAMES = {
    'flux_sensitivity': 'Flux Sensitivity',         # (0, 0) plot position
    'charge_sensitivity': 'Charge Sensitivity',     # (0, 1) plot position
    'anharmonicity': 'Anharmonicity',               # (0, 2) plot position
    't_phi_flux': r'$T_\varphi$ Flux',              # (1, 0) plot position
    't_phi_charge': r'$T_\varphi$ Charge',          # (1, 1) plot position
    't_phi_cc': r'$T_\varphi$ Critical Current',    # (1, 2) plot position
    't1': '$T_1$',                                  # (2, 0) plot position
    't_phi': r'$T_\varphi$',                        # (2, 1) plot position
    'frequency': 'Frequency',                       # (2, 2) plot position
    't2': '$T_2$',                                   # (3, 0) plot position
    'gate_speed': 'Gate Speed',                     # (3, 1) plot position
    'number_of_gates': 'Number of Gates'            # (3, 2) plot position
}

METRIC_KEYS = list(METRIC_NAMES.keys())

# loss keys for plotting.
LOSS_NAMES = {
    'flux_sensitivity_loss': 'Flux Sensitivity Loss',       # (0, 0) plot position
    'charge_sensitivity_loss': 'Charge Sensitivity Loss',   # (0, 1) plot position
    'frequency_loss': 'Frequency Loss',                     # (0, 2) plot position
    'number_of_gates_loss': 'Number of Gates Loss',         # (1, 0) plot position
    'total_loss': 'Total Loss'                              # (1, 1) plot position
}

LOSS_KEYS = list(LOSS_NAMES.keys())

################################################################################
# Helper functions.
################################################################################

def compute_best_ids(
    aggregate_loss_records,
    n: int,
    codes: List[str]
) -> Dict[str, List[int]]:
    out = {}
    for circuit_code in codes:
        code_loss_record = aggregate_loss_records[circuit_code].items()

        sorted_runs = sorted(
            code_loss_record,
            key=lambda x: x[1]['total_loss'][-1]
        )

        out[circuit_code] = [id_num for id_num, run in sorted_runs][:n]

        print(
            f'Top {n} runs in order are '
            f'{out[circuit_code]} for code {circuit_code}.'
        )
    return out


def plot_circuit_metrics(
    run,
    axs,
    plot_type: str,
    circuit_code: str,
    best: bool,
) -> None:

    record_keys = METRIC_KEYS if plot_type == 'metrics' else LOSS_KEYS
    record_names = METRIC_NAMES if plot_type == 'metrics' else LOSS_NAMES
    # l = 3 if plot_type == 'metrics' else 2

    for plot_idx in range(len(record_keys)):
        key = record_keys[plot_idx]
        i, j = plot_idx // 3, plot_idx % 3

        axs[i, j].plot(
            run[key],
            color=PLOT_SCHEME[circuit_code],
            label=circuit_code if best else None,
            alpha=None if best else 0.3,
            linestyle=None if best else '--',
        )
        axs[i, j].set_title(
            record_names[key] + f" {get_units()[key]}"
        )
        axs[i, j].set_yscale('log')
        axs[i, j].legend(loc="upper left")


def plot_results(
    record,
    plot_folder_directory,
    best_ids: Dict[str, List[int]],
    plot_type: str,
    save_suffix: str = '',
) -> None:

    if plot_type == 'metrics':
        fig, axs = plt.subplots(4, 3, figsize=(28, 16))
    elif plot_type == 'loss':
        fig, axs = plt.subplots(2, 3, figsize=(20, 8))
    else:
        raise ValueError(
            f"Unknown plot type: {plot_type}. Must be "
            f"'metrics' or 'loss'"
        )

    for circuit_code, ids_list in best_ids.items():
        # Plot best run for that specific code
        runs_list = [record[circuit_code][id_num] for id_num in ids_list]

        for i, run in enumerate(runs_list):
            # i=0 is always the best circuit.
            plot_circuit_metrics(
                run=run,
                axs=axs,
                plot_type=plot_type,
                circuit_code=circuit_code,
                best=True if i == 0 else False,
            )

    plt.savefig(
        os.path.join(plot_folder_directory, f'{plot_type}_{save_suffix}.png'),
        dpi=500
    )

################################################################################
# Main.
################################################################################


def main() -> None:

    ############################################################################
    # Loading the Yaml file and command line parameters.
    ############################################################################

    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=RECORDS_REQUIRED_KEYS,
        optional_keys=RECORDS_OPTIONAL_KEYS,
    )

    if parameters['num_best'] is not None:
        best_n = int(parameters['num_best'])
    else:
        best_n = int(parameters['num_runs'])

    directory = Directory(parameters, arguments['<yaml_file>'])

    ############################################################################
    # Plotting
    ############################################################################

    circuit_codes = parameters['circuit_code'].split(',')
    aggregate_loss_record = defaultdict(dict)
    aggregate_metrics_record = defaultdict(dict)

    success_count = 0
    for circuit_code in circuit_codes:
        for id_num in range(int(parameters['num_runs'])):

            loss_record = load_record(directory.get_record_file_dir(
                record_type="loss",
                circuit_code=circuit_code,
                idx=id_num,
            ))

            metrics_record = load_record(directory.get_record_file_dir(
                record_type="metrics",
                circuit_code=circuit_code,
                idx=id_num,
            ))

            if loss_record is not None and metrics_record is not None:
                success_count += 1
                aggregate_loss_record[circuit_code][id_num] = loss_record
                aggregate_metrics_record[circuit_code][id_num] = metrics_record

    save_suffix = f'{parameters["circuit_code"]}_{parameters["name"]}_{parameters["optim_type"]}_n_{best_n}'
    best_ids = compute_best_ids(aggregate_loss_record, best_n, circuit_codes)

    plot_results(
        record=aggregate_loss_record,
        plot_folder_directory=directory.get_plots_dir(),
        best_ids=best_ids,
        plot_type='loss',
        save_suffix=save_suffix
    )
    plot_results(
        record=aggregate_metrics_record,
        plot_folder_directory=directory.get_plots_dir(),
        best_ids=best_ids,
        plot_type='metrics',
        save_suffix=save_suffix
    )
    print(
        f"Loaded {success_count} of "
        f"{len(circuit_codes) * int(parameters['num_runs'])} "
        f"successful runs."
    )


if __name__ == "__main__":
    main()



"""
Tabulate results from convergence test.

Usage:
  tabulate.py <yaml_file> [--num_runs=<num_runs>] [--circuit_code=<circuit_code>]
              [--num_best=<num_best>]
  tabulate.py -h | --help
  tabulate.py --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.

Options:
  -h, --help        Show this screen.
  -v, --version     Show version.

  -n, --num_runs=<num_runs>                 Number of runs to consider.
  -c, --circuit_code=<circuit_code>         Circuit code.
  -b, --num_best=<num_best>                 <num_best> number of best.
"""

from collections import defaultdict
import os

from docopt import docopt

from inout import (
    add_command_line_keys,
    Directory,
    load_yaml_file
)

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
YAML_OR_COMMANDLINE_KEYS = [
    "num_runs",
    "circuit_code"
]

################################################################################
# Main.
################################################################################


def main() -> None:

    ############################################################################
    # Loading the Yaml file and command line parameters.
    ############################################################################

    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=YAML_OR_COMMANDLINE_KEYS,
    )

    directory = Directory(parameters, arguments['<yaml_file>'])
    records_dir = directory.get_records_dir()

    ############################################################################
    # Read files
    ############################################################################

    circuit_codes = parameters['circuit_code'].split(',')
    results = defaultdict(lambda:0)

    for circuit_code in circuit_codes:
        for seed in range(int(parameters['num_runs'])):
            save_suffix = (
                f"{circuit_code}"
                f"_{parameters['name']}"
                f"_{seed}"
            )

            summary_save_url = os.path.join(
                records_dir,
                f'summary_{save_suffix}.txt'
            )

            with open(summary_save_url, 'r') as f:
                result = f.readline()
                results[result] += 1
            f.close()

    print("Results:")
    for code in results.keys():
        print(f"Code {code} value: {results[code]}")


if __name__ == "__main__":
    main()



"""Contains helper functions for estimating optimal truncation numbers"""

import logging
from typing import Tuple, List, Optional

import numpy as np
import scipy, scipy.signal
from matplotlib.axes import Axes

from SQcircuit import Circuit, Junction
import SQcircuit.functions as sqf
import SQcircuit.units as unt
from SQcircuit.exceptions import CircuitStateError
from SQcircuit.settings import get_optim_mode


logger = logging.getLogger(__name__)


def get_reshaped_eigvec(
    circuit: Circuit,
    eig_vec_idx: int,
) -> List[np.ndarray]:
    """Returns the component of the ``eig_vec_idx`` eigenvector of ``circuit``
    in each of the modes.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object.
        eig_vec_idx:
            The index of the eigenvector to use.

    Returns
    ----------
        A list of the component of the eigenvector in each of the circuit's
        modes.
    """
    if len(circuit.evecs) == 0:
        raise CircuitStateError('The circuit must be diagonalized first.')

    # Reshape eigenvector dimensions to correspond to individual modes
    if get_optim_mode():
        eigenvector = np.array(circuit.evecs[eig_vec_idx].detach().numpy())
    else:
        eigenvector = circuit.evecs[eig_vec_idx].full()
    eigenvector_reshaped = np.reshape(eigenvector, circuit.m)

    # Get the component of the eigenvector's magnitude in each mode
    mode_magnitudes = []
    total_dim = np.shape(eigenvector)[0]
    for mode_idx, mode_size in enumerate(circuit.m):
        mode_eigvec = np.moveaxis(eigenvector_reshaped, mode_idx, 0)
        mode_eigvec = np.reshape(mode_eigvec, (mode_size, total_dim // mode_size))
        M = np.max(np.abs(mode_eigvec)**2, -1)
        mode_magnitudes.append(M)

    return mode_magnitudes


def fit_mode(
    vector,
    peak_height_threshold=5e-3,
    both_parities=False,
    axis: Optional[Axes]=None,
) -> List[Tuple[float, float, int]]:
    """Given an input vector corresponding to the absolute eigenvector
    magnitudes for a specific harmonic mode, return the decay constant for an
    exponential decay fit starting with the last peak of the input.

    Since eigenstates often have definite parity, when ``both_parities`` is
    ``True``, the fit is done to even/odd indices separately. Otherwise, the
    fit is performed to those indices with the same parity as the maximum
    element in the list.

    Parameters
    ----------
        vector:
            A vector to fit the decay to.
        peak_height_treshold:
            The minimum threshold of an entry to allow to be a peak.
        both_parities:
            Whether to fit to both even and odd indices of ``vector``,
            or only the parity with the maximum element.
        axis:
            An optional matplotlib axis to plot the fit on.

    Returns
    ----------
        A list of tuples ``(decay constant, peak size, peak index)`` for
        each of the parities. 
    """

    def monoExp(x, m, k):
        return m * np.exp(-k * x)

    # Either fit to even and odd indices separately, or just the parity
    # matching the index with the largest value.
    if both_parities:
        parities = [0, 1]
    else:
        parities = [np.argmax(vector) % 2]

    # Fit an exponential decay to each parity
    fit_results = []
    for parity in parities:
        # Get only entries with specified parity
        fit_points_parity = vector[parity::2]

        # Find the peaks in the data
        peaks, _ = scipy.signal.find_peaks(
            fit_points_parity,
            height=peak_height_threshold
        )
        # Edge case: No peaks found, ensure we can use the first entry
        peaks = np.insert(peaks, 0, 0)

        # And get the final one
        fit_peak_idx_parity = sorted(peaks)[-1]
        # Compute the actual index of the peak, to return later
        fit_peak_idx = fit_peak_idx_parity * 2 + parity

        # Fit an exponential decay starting with that peak
        fit_points = fit_points_parity[fit_peak_idx_parity:]

        num_points = len(fit_points)
        fit_range = np.arange(0, 2 * num_points, 2)
        # Choose a smart first guess
        fit_param_guess = (fit_points[0], -np.log(fit_points[-1]) / num_points)
        # And perform the curve fitting
        try:
            params, _ = scipy.optimize.curve_fit(
                monoExp,
                fit_range,
                fit_points,
                fit_param_guess
            )
        except RuntimeError:
            # If cannot fit, just use the guess
            params = fit_param_guess

        m, k = params
        fit_results.append((k, m, fit_peak_idx))

    # Plot slowest decaying curve out of the two parities if `axis` passed.
    if axis:
        if both_parities:
            k, m, peak_idx = get_slow_fit(fit_results)
        else:
            k, m, peak_idx = fit_results[0]
        axis.plot(peak_idx + fit_range, monoExp(fit_range, m, k), 'k')

    return fit_results


def get_slow_fit(
    fit_results: List[Tuple[float, float, int]],
    ignore_threshold=5e-3
) -> Tuple[float, float, float]:
    """Given a list of fit results from ``fit_mode()``, return the one with the
    slowest decay whose peak is at least ``ignore_threshold``.
    
    Parameters
    ----------
        fit_results
            A list of fit results as computed by ``fit_mode``.
        ignore_threshold:
            The minimum size of a peak.

    Returns
    ----------
        The slowest decaying curve whose peak is at least ``ignore_threshold``.
        If all results are below the threshold, it logs a warning and returns
        the first result.
    """
    res_above_threshold = [res for res in fit_results
                           if res[1] >= ignore_threshold]
    if res_above_threshold == []:
        logging.warning('All fit results were below threshold. Returning '
                        'the first result.')
        return fit_results[0]

    # Sort the results in ascending order by decay constant
    res_above_threshold.sort(key = lambda res: res[0])

    # Return result with the smallest constant (slowest-decaying)
    return res_above_threshold[0]


def charge_trunc_to_dimension(trunc: int) -> int:
    """Convert the truncation number for a charge mode to the dimension
    of the Hilbert space.
    
    Given a truncation number ``m`` for a charge mode, the resulting Hilbert
    space dimension is ``2*m - 1``.

    Parameters
    ----------
        trunc:
            The truncation number.

    Returns
    ----------
        The Hilbert space dimension of the corresponding charge mode.
    """

    return 2 * trunc - 1


def dimension_to_charge_trunc(max_dim: int) -> int:
    """Convert the maximum dimension of a Hilbert space to the appropriate
    truncation number for a charge mode.
    
    Given a truncation number ``m`` for a charge mode, the resulting Hilbert
    space dimension is ``2*m - 1``.

    Parameters
    ----------
        max_dim:
            The maximum dimension for the Hilbert space.

    Returns
    ----------
        The largest truncation number such that the resulting Hilbert space
        dimension is no greater than ``max_dim``.
    """

    return int(np.floor((max_dim + 1) / 2))


def charge_mode_heuristic(
    circuit: Circuit,
    n_stds: int = 3
) -> List[int]:
    """Calculate truncation numbers for the charge modes of a circuit by
    approximating each mode as an uncoupled harmonic oscillator and computing
    the width of the (Gaussian) ground state in the number basis.

    Currently only supports a single charge mode.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object to provide truncation numbers for the charge
            modes.
        n_stds:
            The number of standard deviations to include in the truncated
            Hilbert space.

    Returns
    ----------
        List of truncation numbers for each charge mode of ``circuit``.
    """
    charge_modes_idxs = np.flatnonzero(circuit.omega == 0)

    if len(charge_modes_idxs) == 0:
        return []
    if len(charge_modes_idxs) == 1:
        charge_mode_idx = charge_modes_idxs[0]
        EC_eff = (
            (unt.e ** 2) / (2 * np.pi * unt.hbar) / 1e9
            * circuit.cInvTrans[charge_mode_idx, charge_mode_idx]
        )

        EJ_eff = 0
        for _, el, _, w_id in circuit.elem_keys[Junction]:
            EJ = np.squeeze(sqf.to_numpy(el.get_value('GHz'))) / 2 / np.pi
            EJ_eff += EJ * (circuit.wTrans[w_id, charge_mode_idx]) ** 2

        sigma = (EJ_eff / (8 * EC_eff)) ** (1 / 4)
        charge_truncation = int(n_stds * sigma)

        return [charge_truncation]
    else:
        raise NotImplementedError('The charge mode heuristic currently only '
                                  'supports circuits with one charge mode.')


def harmonic_mode_heuristic(
    circuit: Circuit,
    total_trunc_num: int,
    min_trunc_harmonic: int,
    eig_vec_idx: int,
    axes: Optional[Axes]
) -> List[int]:
    """For a diagonalized circuit, suggests a set of truncation numbers for
    harmonic modes that will maximize the likelihood of convergence,
    as defined by ``circuit.check_convergence()``.

    If the circuit has a single harmonic mode, it simply sets the truncation
    number to ``total_trunc_num``.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to provide truncation numbers for.
        eig_vec_idx:
            The index of the eigenvector to use in the heuristic.
        total_trunc_num:
            The maximum Hilbert space dimension for the harmonic mode subspace.
        min_trunc_harmonic:
            The minimum truncation number for harmonic modes to have.
        axes:
            Optionally, a set of Matplotlib Axes to plot the exponential fits
            used in the he

    Returns
    ----------
        List of truncation numbers for the harmonic modes of ``circuit``.
    """
    harmonic_modes = circuit.omega != 0
    n_h = np.count_nonzero(harmonic_modes)

    if n_h == 0:
        return [0]
    elif n_h == 1:
        return [total_trunc_num]

    K = total_trunc_num

    K_eff = K / (min_trunc_harmonic ** n_h)

    # Compute the component of the state in each of the mode
    mode_magnitudes = get_reshaped_eigvec(
        circuit,
        eig_vec_idx,
    )

    decay_constants = np.zeros(n_h, dtype=np.float64)
    peak_locations = np.zeros(n_h, dtype=np.float64)

    # Fit exponential decays to each mode
    for mode_idx in np.flatnonzero(harmonic_modes):
        axis = axes[mode_idx] if axes is not None else None
        fit_results = fit_mode(mode_magnitudes[mode_idx],
                                both_parities=True,
                                axis=axis)
        ki, _, di = get_slow_fit(fit_results)

        decay_constants[mode_idx] = ki
        peak_locations[mode_idx] = di

    # Allocate truncation numbers inverse to decay constants
    harmonic_trunc_nums = (
        np.power(K_eff * np.prod(decay_constants), 1 / n_h)
        / decay_constants
    )

    # Shift by relative peaks
    harmonic_trunc_nums += peak_locations
    # and renormalize to be below K.
    harmonic_trunc_nums *= np.power(K_eff / np.prod(harmonic_trunc_nums),
                                    1 / n_h)

    # Edge case: For mode numbers less than 1, rescale to 1 and rescale
    # other mode numbers to keep total product constant.
    while np.any(harmonic_trunc_nums < 1):
        small_trunc_nums = harmonic_trunc_nums[harmonic_trunc_nums <= 1]
        large_trunc_nums = harmonic_trunc_nums[harmonic_trunc_nums > 1]
        rescale_factor = np.power(np.prod(small_trunc_nums),
                                  1 / len(large_trunc_nums))
        harmonic_trunc_nums[harmonic_trunc_nums > 1] *= rescale_factor
        harmonic_trunc_nums[harmonic_trunc_nums <= 1] = 1

    # Edge case: If a trunc number is greater than K_eff, set it to K_eff
    harmonic_trunc_nums = np.minimum(harmonic_trunc_nums, K_eff)

    # Round to nearest integer and assign harmonic modes. Multiply by
    # `min_trunc_harmonic` because we allocated as a proportion of that.
    harmonic_trunc_nums = np.floor(
        harmonic_trunc_nums * min_trunc_harmonic
    ).astype(int)

    return harmonic_trunc_nums


def trunc_num_heuristic(
    circuit: Circuit,
    eig_vec_idx: int = 0,
    total_trunc_num: int = 1000,
    min_trunc_harmonic: int = 1,
    min_trunc_charge: int = 1,
    max_trunc_charge: Optional[int] = None,
    use_harmonic_heuristic=True,
    use_charge_heuristic=False,
    axes: Optional[Axes] = None
) -> List[int]:
    """For a diagonalized circuits, suggests a set of truncation numbers for
    rediagonalization that will maximize the likelihood of convergence,
    as defined in ``circuit.check_convergence()``.

    If ``use_charge_heuristic`` is ``False``, the truncation numbers for the
    charge modes are set to ``K**(1/circuit.n)``, subject to the values
    of ``min_trunc_charge`` and ``max_trunc_charge``.

    If the circuit has a single mode, it simply sets the truncation number
    to ``total_trunc_num`` (appropriately divided for charge modes).

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to provide truncation numbers for.
        eig_vec_idx:
            The index of the eigenvector to use in the harmonic mode heuristic.
        total_trunc_num:
            The maximum Hilbert space dimension.
        min_trunc_harmonic:
            The minimum truncation number for harmonic modes to have.
        min_trunc_charge:
            The minimum truncation number for charge modes to have. Note a
            truncation number ``m`` for a charge mode results in a Hilbert space
            of dimension ``2*m - 1``.
        max_trunc_charge:
            An optional maximum truncation number for charge modes.
        use_harmonic_heuristic:
            Whether to use the harmonic mode heuristic, or otherwise assign
            the harmonic truncation numbers evenly using the remaining
            Hilbert dimension after assigning the charge modes.  
        use_charge_heuristic:
            Whether to use the charge mode heuristic, or otherwise assign the
            charge mode truncation numbers evenly.
        axes:
            Optionally, a set of Matplotlib Axes to plot the exponential fits
            used in the harmonic mode heuristic.

    Returns
    ----------
        List of truncation numbers for each mode of ``circuit``.
    """
    # Perform checks on parameters
    if len(circuit.evecs) == 0:
        raise CircuitStateError('The circuit must be diagonalized first.')
    if axes is not None:
        if len(axes) < len(circuit.omega):
            raise ValueError(
                'Number of axes for fitting plots should match or exceed '
                'number of modes'
            )
    if total_trunc_num < 1:
        raise ValueError('The maximum dimension of the Hilbert space must '
                         'be an integer >= 1.')
    if min_trunc_harmonic < 1:
        raise ValueError('Harmonic modes must have a minimum truncation number '
                         'of at least 1.')
    if min_trunc_charge  < 1:
        raise ValueError('Charge modes must have a minimum truncation number '
                         'of at least 1.')

    # Deal with the easy case of a single mode (no heuristic necessary)
    if circuit.n == 1:
        logger.info('Single mode (no heuristic required).')
        # Charge mode
        if sum(circuit.omega) == 0:
            return [dimension_to_charge_trunc(total_trunc_num)]
        # Harmonic mode
        else:
            return [total_trunc_num]

    # Set up trunc nums and list of charge, harmonic modes
    trunc_nums = np.zeros_like(circuit.trunc_nums)
    harmonic_modes = circuit.omega != 0
    n_h = np.count_nonzero(harmonic_modes)
    charge_modes = circuit.omega == 0
    n_c = np.count_nonzero(charge_modes)

    # Check that `total_trunc_num` is large enough
    min_dim_req = (
        charge_trunc_to_dimension(min_trunc_charge) ** n_c
        * min_trunc_harmonic ** n_h
    )
    if min_dim_req > total_trunc_num:
        raise ValueError('The `total_trunc_num` passed is not large enough to '
                         'provide each mode the minimum truncation numbers '
                         'requested.')

    # Assign charge mode truncation numbers
    K = total_trunc_num
    if n_c > 0:
        if use_charge_heuristic:
            trunc_nums[charge_modes] = charge_mode_heuristic(circuit)
        else:
            average_dim = K ** (1/len(trunc_nums))
            trunc_nums[charge_modes] = dimension_to_charge_trunc(average_dim)

        # Cut off truncation numbers, if necessary
        if max_trunc_charge is not None:
            trunc_nums[charge_modes & (trunc_nums > max_trunc_charge)] = max_trunc_charge

        trunc_nums[charge_modes & (trunc_nums < min_trunc_charge)] = min_trunc_charge

        # Compute the Hilbert space dimension remaining for harmonic modes
        charge_space_dim = np.prod(charge_trunc_to_dimension(trunc_nums[charge_modes]))
        K //= charge_space_dim

        if K < (min_trunc_harmonic ** n_h):
            raise ValueError(
                'The charge truncation numbers were too large to provide each '
                f'harmonic mode a truncation number >={min_trunc_harmonic}. '
                'Please pass a smaller value for `max_trunc_charge`.'
            )

    # Assign harmonic mode truncation numbers
    if n_h > 0:
        if use_harmonic_heuristic:
            harmonic_trunc_nums = harmonic_mode_heuristic(
                circuit,
                K,
                min_trunc_harmonic,
                eig_vec_idx,
                axes
            )
        else:
            harmonic_trunc_nums = [np.floor(K**(1/n_h)).astype(int)] * n_h

        # Because everything is integers, it is possible in the process of
        # taking the floor we cut off too much. Thus, now maximize each
        # individual truncation number while ensuring product less than K.
        for idx in range(len(harmonic_trunc_nums)):
            curr_dim = np.prod(harmonic_trunc_nums)
            if curr_dim == K:
                break

            harmonic_trunc_nums[idx] += np.floor(
                (K - curr_dim) 
                * harmonic_trunc_nums[idx]
                / curr_dim,
            ).astype(int)

        trunc_nums[harmonic_modes] = harmonic_trunc_nums

    return list(trunc_nums)


def assign_trunc_nums(
    circuit: Circuit,
    total_trunc_num: int,
    min_trunc_harmonic: int = 1,
    min_trunc_charge: int = 1,
    max_trunc_charge: Optional[int] = None,
    use_charge_heuristic=False,
    axes: Optional[Axes] = None
) -> List[int]:
    """Heuristically re-assign truncation numbers for a circuit with multiple
    modes. (In the case of a single mode, it simply assigns the truncation
    number to ``total_trunc_num``.)

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to provide truncation numbers for.
        total_trunc_num:
            The maximum Hilbert space dimension.
        min_trunc_harmonic:
            The minimum truncation number for harmonic modes.
        min_trunc_charge:
            The minimum truncation number for charge modes.
        max_trunc_charge:
            An optional maximum truncation number for charge modes.
        use_charge_heuristic:
            Whether to use the charge mode heuristic, or just assign the
            charge mode truncation numbers evenly.
        axes:
            Optionally, a set of Matplotlib Axes to plot the exponential fits
            during the charge mode heuristic.

    Returns
    ----------
        trunc_nums:
            List of truncation numbers for each mode of ``circuit``.
    """
    logger.info('Reallocating truncation numbers.')
    trunc_nums = trunc_num_heuristic(
            circuit,
            total_trunc_num=total_trunc_num,
            eig_vec_idx=1,
            min_trunc_harmonic=min_trunc_harmonic,
            min_trunc_charge=min_trunc_charge,
            max_trunc_charge=max_trunc_charge,
            use_charge_heuristic=use_charge_heuristic,
            axes=axes
    )

    circuit.set_trunc_nums(trunc_nums)

    return circuit.trunc_nums


def test_convergence(
    circuit: Circuit,
    eig_vec_idx: int = 0,
    t: int = 10,
    threshold: float = 1e-5,
    test_charge_modes = False
) -> Tuple[bool, List[float]]:
    """Test convergence of a circuit.

    Requires the last ``t`` (if available) entries corresponding to
    each mode individually of the ``eig_vec_idx`` eigenvector are on
    average less than ``threshold``.

    By default this does **not** test charge modes, since this convergence test
    is written with harmonic modes (exponentially decaying magnitudes) in mind.

    Parameters
    ----------
        circuit:
            Circuit to test convergence of. Must be diagonalized.
        eig_vec_idx:
            The index of the eigenvector to use to check convergence.
        t:
            The number of entries of the eigenvector to use to check
            convergence. Must be at least 2.
        threshold:
            The maximum for the last ``t`` entries of the eigenvector.
        test_charge_modes:
            Whether to test the charge modes (False by default).

    Returns
    ----------
        is_converged:
            A boolean of whether the circuit converged.
        epsilons:
            The average values of the last ``t`` components in each mode.
    """
    if len(circuit.efreqs) == 0:
        raise CircuitStateError('The circuit must be diagonalized first.')
    if np.any(np.array(circuit.m) < 4):
        raise ValueError('In order to check both parities, each mode must '
                         'have a dimension >= 4.')
    if t < 2:
        raise ValueError('The value for `t` must be at least 2.')

    mode_magnitudes = get_reshaped_eigvec(
        circuit,
        eig_vec_idx
    )

    # Get the last `t` entries of the component of the eigenvector in each mode.
    y = [M[-t:] for M in mode_magnitudes]

    # Average them. If there are fewer than `t` entires, only average the last
    # two (to avoid averaging the whole component).
    epsilons = []
    for yi in y:
        if len(yi) <= t:
            epsilon_i = np.average(yi[-2:])
        else:
            epsilon_i = np.average(yi)
        epsilons.append(epsilon_i)
    # epsilons = [np.average(yi) for yi in y]

    for mode_idx, epsilon_i in enumerate(epsilons):
        if epsilon_i > threshold:
            if circuit.omega[mode_idx] == 0 and not test_charge_modes:
                continue
            return False, epsilons

    return True, epsilons



"""This module implements the SGD optimizer."""

from typing import Dict, Optional

from SQcircuit import Circuit, CircuitComponent
from torch.optim import SGD
from torch.optim.lr_scheduler import ExponentialLR
from torch import Tensor

from .optim import OptimizationRecord, run_optimization
from .utils import LossFunctionType

DEFAULT_OPTIM_KWARGS = {
    'nesterov': False,
    'momentum': 0.0,
    'lr': 1e-2
}

SCHEDULER_KWARGS = {
    'gamma': 0.99
}

def run_SGD(
    circuit: Circuit,
    loss_metric_function: LossFunctionType,
    max_iter: int,
    total_trunc_num: int,
    bounds: Dict[CircuitComponent, Tensor],
    optimizer_kwargs: Optional[Dict] = None,
    num_eigenvalues: int = 10,
    save_loc: Optional[str] = None,
    identifier: Optional[str] = None,
    save_intermediate_circuits: bool = False,
) -> OptimizationRecord:
    """Runs SGD for a maximum of ``max_iter`` beginning with ``circuit`` using
    ``loss_metric_function``.

    Parameters
    ----------
        circuit:
            A circuit to optimize.
        loss_metric_function:
            Loss function to optimize.
        max_iter:
            Maximum number of iterations.
        total_trunc_num:
            Maximum total truncation number to allocate.
        bounds:
            Dictionary giving bounds for each element type.
        num_eigenvalues:
            Number of eigenvalues to calculate when diagonalizing.
        save_loc:
            Folder to save results in, or None.
        identifier:
            String identifying this run (e.g. seed, name, circuit code, …)
            to use when saving.
        save_intermediate_circuits:
            Whether to save the circuit at each iteration.

    Returns
    ----------
        An ``OptimizationRecord`` containing the optimized circuit, the final
        loss, and a record of the component loss and metric values at each
        epoch.   
    """

    # Set up optimizer kwargs
    if optimizer_kwargs is None:
        optimizer_kwargs = {}

    optimizer_kwargs = DEFAULT_OPTIM_KWARGS | optimizer_kwargs

    return run_optimization(
        circuit = circuit,
        loss_metric_function = loss_metric_function,
        max_iter = max_iter,
        total_trunc_num = total_trunc_num,
        bounds = bounds,
        optimizer = SGD,
        uses_closure = False,
        optimizer_kwargs = optimizer_kwargs,
        scheduler=ExponentialLR,
        scheduler_kwargs=SCHEDULER_KWARGS,
        num_eigenvalues = num_eigenvalues,
        save_loc = save_loc,
        identifier = identifier,
        save_intermediate_circuits = save_intermediate_circuits
    )



"""Contains helper functions used in remainder of code."""
from copy import copy
import logging
from typing import Callable, Tuple

import numpy as np
import torch
from SQcircuit import Circuit
from SQcircuit.settings import get_optim_mode
from SQcircuit.units import get_unit_freq

from .utils import (
    construct_perturbed_elements,
    zero,
    SQValType
)

logger = logging.getLogger(__name__)

# NOTE: All functions must treat the input `circuit` as const, at least
# in effect.

def first_resonant_frequency(circuit: Circuit) -> SQValType:
    """Calculates resonant frequency of first excited eigenstate in circuit.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.

    Returns
    ----------
        The frequency f_{10}, in the frequency units of ``SQcircuit``.
    """
    omega = circuit.efreqs[1] - circuit.efreqs[0]
    return omega


def fastest_gate_speed(circuit: Circuit) -> SQValType:
    """Calculates the upper bound for the speed of the single-qubit gate.
    The upper bound is:
    min{f_i0 - f_10, |f_i0 - 2f_10|}; for i>1

    Parameters
    ----------
        circuit:
            A ``Circuit`` object specifying the qubit.

    Returns
    ----------
        The uppper bound on a single-qubit gate, in the frequency units of
        ``SQcircuit``.
    """

    omega = circuit.efreqs[2] - circuit.efreqs[1]

    for i in range(2, len(circuit.efreqs)):

        f_i1 = circuit.efreqs[i] - circuit.efreqs[1]
        anharm_i = abs(
            (circuit.efreqs[i] - circuit.efreqs[0]) -
            2*(circuit.efreqs[1] - circuit.efreqs[0])
        )

        if f_i1 < omega:
            omega = f_i1
        elif anharm_i < omega:
            omega = anharm_i

    return omega


def decoherence_time(circuit: Circuit, t_type: str, dec_type: str) -> SQValType:
    """Return the decoherence time for a given circuit and decoherence type,
    between the 0th and 1st eigenstates.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to calculate the decoherence time of.
        t_type:
            A string specifying the type of decoherence. It must be either
            ``'t1'`` or ``'t_phi'``.
        dec_type:
            A string specifying the channel of the decoherence time. It must be
            either ``'capacitive'``, ``'inductive'``, ``'quasiparticle'``,
            ``'charge'``, ``'cc'``, ``'flux'``, or ``'total'``.

    Returns
    ----------
        The decoherence time of ``circuit`` of type ``t_type`` via channel
        ``dec_type``.
    """

    gamma = zero()

    if t_type == 't1':
        all_t1_channels = ['capacitive', 'inductive', 'quasiparticle']
        if dec_type == 'total':
            dec_type_list = all_t1_channels
        else:
            if dec_type not in all_t1_channels:
                raise ValueError(
                    "dec_type with 't1' mode should be {all_t1_channels}, or "
                    "'total'"
                )
            dec_type_list = [dec_type]
    elif t_type == 't_phi':
        all_tp_channels = ['charge', 'cc', 'flux']
        if dec_type == 'total':
            dec_type_list = all_tp_channels
        else:
            if dec_type not in all_tp_channels:
                raise ValueError(
                    "dec_type with 't_phi' mode should be in {all_tp_channels}"
                    ", or 'total'"
                )
            dec_type_list = [dec_type]
    else:
        raise ValueError("t_type must be either 't1' or 't_phi'")

    for dec_type in dec_type_list:
        gamma = gamma + circuit.dec_rate(dec_type, (0, 1))

    return 1 / gamma


def total_dec_time(circuit: Circuit) -> SQValType:
    """Return the T2 (total decohence) time.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to calculat the T2 of.

    Returns
    ----------
        The T2 time of ``circuit``, between the 0th and 1st eigenstates,
        via the capacitive, inductive, quasiparticle, charge, critical current,
        and flux channels.
    """
    t1 = decoherence_time(
        circuit=circuit,
        t_type='t1',
        dec_type='total'
    )

    tp = decoherence_time(
        circuit=circuit,
        t_type='t_phi',
        dec_type='total'
    )

    return 2*t1*tp / (2*t1+tp)


def number_of_gates(circuit: Circuit) -> SQValType:
    """Return the number of single-qubit gates, by calculating the gate speed
    (loosely, an anharmonicity) and total decoherence time.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object.
        
    Returns
    ----------
        The maximum number of sequential single-qubit gates which can be
        executed.
    """

    # don't forget the units!
    gate_speed = fastest_gate_speed(circuit) * get_unit_freq()

    t = total_dec_time(circuit)

    return gate_speed * t


def charge_sensitivity(
    circuit: Circuit,
    n_samples: int = 4,
    epsilon: float = 1e-14,
) -> SQValType:
    """Compute the sensitivity in the circuit's qubit frequency to drift in
    gate charge in all charge islands. To compute this, samples the qubit
    frequency 'diagonally' in charge space: evenly samples ``n_samples`` points
    along the diagonal line n_{g1} = n_{g2} = ..., computes the qubit frequency
    at each point, and returns a normalized range of variation (max frequency
    sampled minus min, over the average frequency).
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object to calculate the sensitivity of.
        n_samples:
            The number of gate charge points to sample.
        epsilon:
            The value to return if there are no charge modes (so the
            sensitivity is formally zero).

    Returns
    ----------
        The normalized range of f_{10} across all gate charge points sampled.
    """

    charge_offsets = np.linspace(0.0, 0.5, n_samples)
    efreq_values = []

    # Edge case: For circuit with no charge modes, assign zero sensitivity
    if np.all(circuit.omega != 0):
        if get_optim_mode():
            return torch.as_tensor(epsilon) + 0 * circuit.efreqs[0]
        else:
            return epsilon

    # Assume all charge modes set to 0 initially
    freq_q = circuit.efreqs[1] - circuit.efreqs[0]
    efreq_values.append(freq_q)

    # Copy circuit to create new container for perturbed eigenstates
    perturb_circ = copy(circuit)
    # For each mode, if charge mode exists, then set gate charge to obtain
    # minimum frequency

    for charge_offset in charge_offsets:
        # we have already added this to the efreqs_values
        if charge_offset == 0.0:
            continue

        for charge_island_idx in perturb_circ.charge_islands.keys():
            perturb_circ.set_charge_offset(
                charge_island_idx+1,
                charge_offset,
            )

        perturb_circ.diag(len(circuit.efreqs))
        freq_q = perturb_circ.efreqs[1] - perturb_circ.efreqs[0]
        efreq_values.append(freq_q)

    # Reset charge modes to 0; this is necessary because perturb_circ and
    # circ are basically the same
    for charge_island_idx in circuit.charge_islands.keys():
        perturb_circ.set_charge_offset(charge_island_idx + 1, 0.)

    if get_optim_mode():
        c_max = torch.max(torch.stack(efreq_values))
        c_min = torch.min(torch.stack(efreq_values))
        return 2 * (c_max - c_min) / (c_max + c_min)
    else:
        c_max = np.max(np.stack(efreq_values))
        c_min = np.min(np.stack(efreq_values))
        return 2 * (c_max - c_min) / (c_max + c_min)


def flux_sensitivity(
    circuit: Circuit,
    delta=0.01,
    loop_idx=0,
) -> SQValType:
    """Compute the sensitivity in the circuit's qubit frequency to variations
    in external flux through one of its loops. The flux sensitivity is defined
    as the normalized change in qubit frequency for a perturbation of ``delta``
    about the current external flux, i.e.
        ``|f_{10}(phi_ext + delta) - f_{10}(phi_ext)|/f_{10}(phi_ext)``. 

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to calculate the sensitivity of.
        delta:
            The perturbation in internal flux.
        loop_idx:
            The index of the loop in ``circuit.loops`` to perturb.

    Returns
    ----------
        The normalized variation of f_{10} to a ``delta`` perturbation in
        external flux through the ``loop_idx``th loop.
    """
    f_0 = circuit.efreqs[1] - circuit.efreqs[0]

    # Copy circuit to create new container for perturbed eigenstates
    perturb_circ = copy(circuit)
    loop = perturb_circ.loops[loop_idx]
    org_flux = loop.internal_value

    # Change the flux and get the eigen-frequencies
    loop.internal_value = org_flux + delta * 2 * np.pi
    perturb_circ.diag(len(circuit.efreqs))
    f_delta = perturb_circ.efreqs[1] - perturb_circ.efreqs[0]

    # Return loop back to original flux
    loop.internal_value = org_flux

    if get_optim_mode():
        S = torch.abs((f_delta - f_0) / f_0)
    else:
        S = np.abs((f_delta - f_0) / f_0)

    return S


def element_sensitivity(
    circuit: Circuit,
    circuit_metric: Callable[[Circuit], SQValType],
    n_samples=25,
    fabrication_error=0.01,
) -> Tuple[SQValType, SQValType]:
    """Compute an estimate of circuit sensitivity, as determined by the
    standard deviation in a circuit metric calculated for circuits with a
    normally-distributed fabrication error.

    In particular, constructs ``n_samples`` circuits with elements which are
    sampled according to a normal distribution about the values in ``circuit``,
    assuming a percentage fabrication error of ``fabrication_error``. Computes
    ``circuit_metric`` for each sampled circuit and returns the normalized
    standard deviation (std over mean) of values calculated.
    
    Only works when ``SQcircuit``'s optimization mode is on (assumes all
    element values are Tensors).

    Parameters
    ----------
        circuit:
            A ``Circuit`` object to use as the center of the distribution.
        circuit_metric:
            A metric to calculate the sensitivity of ``circuit`` to. 
        n_samples:
            The number of randomly sampled circuits to calculate. More samples
            provide more reproducible results.
        fabrication_error:
            The percentage fabrication error to simulate. The elements for the
            new circuit are sampled from a normal distribution with standard
            deviation ``fabrication_error * element_value``.

    Returns
    ----------
        The normalized standard deviation of ``circuit_metric`` over
        ``n_samples`` randomly sampled circuits.
    """
    dist = torch.distributions.MultivariateNormal(
        torch.stack(circuit.parameters),
        torch.diag((fabrication_error * torch.stack(circuit.parameters)) ** 2)
    )
    # re-parameterization trick
    new_params = dist.rsample((n_samples, ))
    vals = torch.zeros((n_samples,))

    logger.info('Calculating element sensitivity...')
    for i in range(n_samples):
        logger.info('%i/%i', i+1, n_samples)
        elements_sampled = construct_perturbed_elements(circuit,
                                                        new_params[i,:])
        cr_sampled = Circuit(elements_sampled)
        cr_sampled.set_trunc_nums(circuit.trunc_nums)
        cr_sampled.diag(len(circuit.efreqs))
        vals[i] = circuit_metric(cr_sampled)

    sensitivity = torch.std(vals) / torch.mean(vals)

    return sensitivity


def reset_charge_modes(circuit: Circuit) -> None:
    """Sets gate charge of all charge degrees of freedom to zero.
    
    Parameters
    ----------
        circuit:
            A ``Circuit`` object to reset.
    """
    default_n_g = 0.0

    if sum(circuit.omega == 0) == 0:
        return
    else:
        for charge_island_idx in circuit.charge_islands.keys():
            charge_mode = charge_island_idx + 1
            circuit.set_charge_offset(charge_mode, default_n_g)



"""
Plot flux/charge spectra and computational subspace phase plots for a given set
of circuits.

Usage:
  plot_analysis.py <yaml_file> [--ids=<ids>] [--circuit_code=<circuit_code>]
                   [--optim_type=<optim_type>]
  plot_analysis.py -h | --help
  plot_analysis.py --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.

Options:
  -h, --help    Show this screen.
  --version     Show version.

  -i, --ids=<ids>                           ids of the circuits to be summarized
  -c, --circuit_code=<circuit_code>         Circuit code
  -o, --optim_type=<optim_type>             Optimization method
"""

import os

from docopt import docopt
import matplotlib.pyplot as plt
import numpy as np
import SQcircuit as sq

import analysis as an
from utils import load_final_circuit, set_plotting_defaults
from inout import (
    add_command_line_keys,
    Directory,
    load_yaml_file
)

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
YAML_OR_COMMANDLINE_KEYS = [
    "ids",
    "circuit_code",
    "optim_type"
]


def main() -> None:

    ############################################################################
    # Load the Yaml file and command line parameters.
    ############################################################################
    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=YAML_OR_COMMANDLINE_KEYS,
    )

    name = parameters['name']
    circuit_code = parameters['circuit_code']
    optim_type = parameters['optim_type']
    ids = parameters['ids'].split(',')

    directory = Directory(parameters, arguments['<yaml_file>'])
    plot_output_folder = directory.get_plots_dir()

    ############################################################################
    # Plotting
    ############################################################################

    set_plotting_defaults()

    sq.set_engine('PyTorch')
    for id_num in ids:
        circuit_path = directory.get_record_file_dir(
            record_type='circuit',
            circuit_code=circuit_code,
            idx=id_num,
        )
        cr = load_final_circuit(circuit_path)
        cr.update()  # rebuild op memory

        save_prefix = f'plot_{circuit_code}_{name}_{optim_type}_{id_num}'
        # Plot flux
        fig, ax = plt.subplots()
        flux_vals, flux_spectra = an.calculate_flux_spectrum(cr)
        an.plot_flux_spectrum(flux_vals, flux_spectra, ax)
        save_path = os.path.join(plot_output_folder, f'{save_prefix}.flux.png')
        plt.savefig(
            save_path,
            bbox_inches="tight",
            dpi=300
        )
        ax.clear()

        # Plot charge
        if np.count_nonzero(cr.omega == 0) > 0:
            # Plot diagonal
            ng_vals, charge_spectra = an.sweep_charge_spectrum(
                cr,
                [True] * len(cr.m)
            )
            an.plot_1d_charge_spectrum(ng_vals, charge_spectra, ax)
            save_path = os.path.join(
                plot_output_folder, f'{save_prefix}.charge_diag.png'
            )
            plt.savefig(
                save_path,
                dpi=300,
                bbox_inches="tight"
            )
            ax.clear()


if __name__ == '__main__':
    main()



"""set of utility functions to apply a reparameterization to circuit element
values. We call the reparameterization the "alpha" parameters.
"""

from typing import Dict, Type

from SQcircuit import Circuit, CircuitComponent, Loop
import torch
from torch import Tensor

def get_alpha_param_from_circuit_param(
    circuit_param: Tensor,
    bounds: Dict[CircuitComponent, Tensor],
    elem_type: Type[CircuitComponent]
) -> Tensor:
    """Compute the alpha parameter corresponding to a circuit parameter.

    Parameters
    ----------
        circuit_param:
            A circuit parameter (element value).
        bounds:
            A dictionary of the element bounds to use to compute the alpha
            parameter.
        elem_type:
            The element type that ``circuit_param`` corresponds to.

    Returns
    ----------
        The alpha parameter corresponding to ``circuit_param``.
    """
    l_bound, u_bound = bounds[elem_type]

    if elem_type == Loop:
        var = (u_bound - l_bound) / 2
        mean = (u_bound + l_bound) / 2

        return torch.acos((circuit_param - mean) / var)
    else:
        var = (torch.log(u_bound) - torch.log(l_bound))/2
        mean = (torch.log(u_bound) + torch.log(l_bound))/2

        return torch.acos((torch.log(circuit_param) - mean) / var)


def get_alpha_params_from_circuit_params(
    circuit: Circuit,
    bounds: Dict[CircuitComponent, Tensor]
) -> Tensor:
    """Compute the alpha parameters for a circuit.

    Parameters
    ----------
        circuit:
            A ``Circuit`` object.
        bounds:
            A dictionary of the element bounds to use to compute the alpha
            parameters.

    Returns
    ----------
        The alpha parameters corresponding to ``circuit.parameters``.
    """
    alpha_params = torch.zeros(torch.stack(circuit.parameters).shape,
                               dtype=torch.float64)

    for param_idx, circuit_param in enumerate(circuit.parameters):
        alpha_params[param_idx] = get_alpha_param_from_circuit_param(
            circuit_param,
            bounds,
            circuit.get_params_type()[param_idx]
        )
    return alpha_params


def get_circuit_param_from_alpha_param(
    alpha_param: Tensor,
    bounds: Dict[CircuitComponent, Tensor],
    elem_type: Type[CircuitComponent],
) -> Tensor:
    """Compute the circuit parameter corresponding to an alpha parameter.

    Parameters
    ----------
        alpha_param:
            The alpha parameter.
        bounds:
            A dictionary of the element bounds used when computing
            ``alpha_param``.
        elem_type:
            The type of element corresponding to ``alpha_param``.

    Returns
    ----------
        The circuit parameter (element value) corresponding to ``alpha_param``.
    """
    l_bound, u_bound = bounds[elem_type]

    if elem_type == Loop:
        var = (u_bound - l_bound) / 2
        mean = (u_bound + l_bound) / 2

        return mean + var * torch.cos(alpha_param)
    else:
        var = (torch.log(u_bound) - torch.log(l_bound))/2
        mean = (torch.log(u_bound) + torch.log(l_bound))/2

        return torch.exp(mean + var * torch.cos(alpha_param))


def get_circuit_params_from_alpha_params(
    alpha_params: Tensor,
    circuit: Circuit,
    bounds: Dict[CircuitComponent, Tensor]
) -> Tensor:
    """Compute the parameters for a circuit corresponding to alpha 
    parameters.
    
    Parameters
    ----------
        alpha_params:
            A Tensor of alpha_params, corresponding to the parameters of
            ``circuit``.
        circuit:
            A ``Circuit`` object.
        bounds:
            A dictionary of the element bounds used when computing
            ``alpha_params``.

    Returns
    ----------
        A Tensor of circuit parameters corresponding to ``alpha_params``.
    """
    circuit_params = torch.zeros(alpha_params.shape, dtype=torch.float64)

    for param_idx, alpha_param in enumerate(alpha_params):
        circuit_params[param_idx] = get_circuit_param_from_alpha_param(
            alpha_param,
            bounds,
            circuit.get_params_type()[param_idx]
        )
    return circuit_params



import numpy as np

import SQcircuit as sq
import SQcircuit.functions as sqf
from SQcircuit.settings import get_optim_mode


################################################################################
# Initialize circuit
################################################################################

def reset_loop_flux(circuit, flux):
    for loop in circuit.loops:
        loop.set_flux(flux)


def reset_charge_modes(circuit, default_charge=0):
    for charge_island_idx in circuit.charge_islands.keys():
        charge_mode = charge_island_idx + 1
        circuit.set_charge_offset(charge_mode, default_charge)

################################################################################
# Print element values
################################################################################


def round_n_sigfigs(value):
    scale = np.floor(np.log10(value))
    value *= np.power(10, -scale)
    value = np.round(value, 3)
    value /= np.power(10, -scale)
    return value


def get_default_unit(element):
    if type(element) is sq.Junction:
        return 'GHz'
    elif type(element) is sq.Capacitor:
        return 'F'
    elif type(element) is sq.Inductor:
        return 'H'
    return ''


def build_circuit_topology_string(circuit):
    output = "\n"
    for circuit_edge, elements in circuit.elements.items():
        output += f"Edge {circuit_edge}\n"

        for element_idx, element in enumerate(elements):
            output += f"\tElement {element_idx}: {type(element)}\n"
            element_unit = get_default_unit(element)
            element_value = element.get_value(element_unit)
            if type(element) is sq.Junction:
                element_value /= (2 * np.pi)
            if get_optim_mode():
                element_value = element_value.detach().numpy()
            element_value = round_n_sigfigs(element_value)
            output += f"\tValue: {element_value} {element_unit}\n"
        output += "\n"
    return output


################################################################################
# Plot charge spectrum
################################################################################
def sweep_charge_spectrum(
    circuit,
    sweep_charge_node_bools,
    n_eig=4,
    charge_min=0,
    charge_max=1,
    default_charge=0,
    n_charge_points=20
):
    circuit_initial_flux = circuit.loops[0].value() / (2 * np.pi)
    reset_loop_flux(circuit, circuit_initial_flux)

    n_g_vals = np.linspace(charge_min, charge_max, n_charge_points)
    spectrum = np.zeros((n_eig, n_charge_points))

    if sum(circuit.omega == 0) == 0:
        return None
    else:
        print("Calculating charge spectrum...")
        for charge_sweep_idx, n_g in enumerate(n_g_vals):
            print(f"{charge_sweep_idx+1}/{n_charge_points}")
            for charge_island_idx in circuit.charge_islands.keys():
                if sweep_charge_node_bools[charge_island_idx]:
                    circuit.set_charge_offset(
                        charge_island_idx + 1, n_g
                    )
                else:
                    circuit.set_charge_offset(
                        charge_island_idx + 1, default_charge
                    )

            # circuit.update() # necessary?
            x, _ = circuit.diag(n_eig)
            spectrum[:, charge_sweep_idx] = sqf.to_numpy(x)

    # Set zeroth eigenfrequency to zero
    spectrum -= spectrum[0, :]
    return n_g_vals, spectrum


def plot_1d_charge_spectrum(n_g_vals, spectrum, ax, n_eig=4):
    for eigen_idx in range(n_eig):
        ax.plot(n_g_vals, spectrum[eigen_idx, :])

    ax.set_xlabel(r"$n_g$")
    ax.set_ylabel(r"$f_i-f_0$ [GHz]")

################################################################################
# Plot flux spectrum
################################################################################


def calculate_flux_spectrum(
    circuit,
    flux_min=0,
    flux_max=0.5,
    n_eig=4,
    count=30,
):
    reset_charge_modes(circuit)

    # Currently assume one loop
    initial_flux_value = circuit.loops[0].value() / (2 * np.pi)
    flux_values = np.linspace(flux_min, flux_max, count)
    flux_spectrum = np.zeros((n_eig, count))
    print("Calculating flux spectrum...")
    for flux_idx, flux in enumerate(flux_values):
        print(f"{flux_idx+1}/{count}")
        for loop in circuit.loops:
            loop.set_flux(flux)
        circuit.update()
        eigenvalues, _ = circuit.diag(n_eig)
        flux_spectrum[:, flux_idx] = sqf.to_numpy(eigenvalues)
        flux_spectrum -= flux_spectrum[0, :]

    # Reset circuit
    for loop in circuit.loops:
        loop.set_flux(initial_flux_value)
    circuit.update()
    circuit.diag(n_eig)
    return flux_values, flux_spectrum


def plot_flux_spectrum(flux_vals, spectrum, ax, n_eig=-1):
    if n_eig == -1:
        n_eig = np.shape(spectrum)[0]
    for eigen_idx in range(n_eig):
        ax.plot(flux_vals, spectrum[eigen_idx, :])

    ax.set_xlabel(r"$\Phi_{ext}/\Phi_0$")
    ax.set_ylabel(r"$f_i-f_0$ [GHz]")



"""test_sampler contains tests for generating random circuit topologies with
circuit values sampled from a fixed range."""

import SQcircuit as sq
import numpy as np

from SQcircuit import Circuit
from SQcircuit.settings import set_optim_mode

from qubit_discovery import CircuitSampler


def test_convergence():
    # Create fluxonium circuit
    def create_fluxonium(
        cap_value, cap_unit, ind_value, ind_unit,
        junction_value, junction_unit
    ):
        loop = sq.Loop()
        loop.set_flux(0)

        capacitor = sq.Capacitor(cap_value, cap_unit, Q=1e6)
        inductor = sq.Inductor(ind_value, ind_unit, Q=500e6, loops=[loop])
        junction = sq.Junction(
            junction_value, junction_unit, cap=capacitor,
            A=1e-7, x=3e-06, loops=[loop]
        )
        circuit_fluxonium = sq.Circuit(
            {(0, 1): [capacitor, inductor, junction], }
        )
        return circuit_fluxonium
    trunc_cutoff = 100
    trunc_range = np.arange(12, trunc_cutoff + 1)
    cutoff = 3

    num_eigenvalues = 10
    for x in trunc_range:
        fluxonium = create_fluxonium(
            2, 'GHz', 0.46,
            'GHz', 10.2, 'GHz'
        )
        # Assuming contiguous divergence/convergence trunc nums, find the lowest
        # truncation number that converges
        fluxonium.set_trunc_nums([x, ])
        fluxonium.diag(num_eigenvalues)
        if not fluxonium.check_convergence(threshold=1e-5)[0]:
            cutoff = x + 1
    assert cutoff == 41


def get_elements_from_code(circuit_code: str) -> Circuit:

    sampler = CircuitSampler(
        capacitor_range=[12e-15, 12e-9],
        inductor_range=[12e-9, 12e-6],
        junction_range=[1e9, 10e9],
        flux_range=[0.5, 0.5],
        elems_not_to_optimize=[sq.Loop]
    )
    circuit = sampler.sample_circuit_code(circuit_code)

    return circuit.elements


def test_circuit_sampling_from_code_1() -> None:

    elements = get_elements_from_code("JL(JC)")

    elem1, elem2, elem3 = elements[(0, 1)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Inductor)
    assert isinstance(elem3, sq.Capacitor)
    assert elem1.loops != []
    assert elem2.loops != []

    elem1, elem2 = elements[(1, 2)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Capacitor)
    assert elem1.loops == []

    elem, = elements[(2, 0)]
    assert isinstance(elem, sq.Capacitor)


def test_circuit_sampling_from_code_2() -> None:

    elements = get_elements_from_code("J(JC)J(LC)")

    elem1, elem2, elem3 = elements[(0, 1)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Junction)
    assert isinstance(elem3, sq.Capacitor)
    assert elem1.loops != []
    assert elem2.loops != []

    elem1, elem2 = elements[(0, 2)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Capacitor)
    assert elem1.loops == []

    elem, = elements[(2, 1)]
    assert isinstance(elem, sq.Capacitor)

    elem1, elem2 = elements[(1, 3)]
    assert isinstance(elem1, sq.Inductor)
    assert isinstance(elem2, sq.Capacitor)
    assert elem1.loops == []

    elem, = elements[(3, 0)]
    assert isinstance(elem, sq.Capacitor)

    elem, = elements[(2, 3)]
    assert isinstance(elem, sq.Capacitor)


def test_circuit_sampling_from_code_3() -> None:

    elements = get_elements_from_code("JJ(J(LC)C)")

    elem1, elem2, elem3 = elements[(0, 1)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Junction)
    assert isinstance(elem3, sq.Capacitor)
    assert elem1.loops != []
    assert elem2.loops != []

    elem1, elem2 = elements[(1, 2)]
    assert isinstance(elem1, sq.Junction)
    assert isinstance(elem2, sq.Capacitor)
    assert elem1.loops == []

    elem1, elem2 = elements[(1, 3)]
    assert isinstance(elem1, sq.Inductor)
    assert isinstance(elem2, sq.Capacitor)

    elem, = elements[(3, 2)]
    assert isinstance(elem, sq.Capacitor)

    elem, = elements[(2, 0)]
    assert isinstance(elem, sq.Capacitor)

    elem, = elements[(0, 3)]
    assert isinstance(elem, sq.Capacitor)


def test_circuit_sampling_from_code_4() -> None:

    set_optim_mode(True)

    elements = get_elements_from_code("flux_qubit")

    junc_1, cap_1 = elements[(0, 1)]
    junc_2, cap_2 = elements[(1, 2)]
    junc_3, cap_3 = elements[(2, 0)]

    assert junc_1 == junc_2
    assert cap_1 == cap_2
    assert junc_2 != junc_3
    assert cap_2 != cap_3

    circuit = sq.Circuit(elements, flux_dist="junctions")

    assert len(circuit.parameters) == 4

    set_optim_mode(False)


def test_circuit_sampling_from_code_5() -> None:

    set_optim_mode(True)

    elements = get_elements_from_code("transmon")

    junc_1, junc_2, cap = elements[(0, 1)]

    assert junc_1 == junc_2

    circuit = sq.Circuit(elements, flux_dist="junctions")

    circuit.description()

    assert len(circuit.parameters) == 2

    set_optim_mode(False)


def test_circuit_sampling_from_code_6() -> None:

    set_optim_mode(True)

    elements = get_elements_from_code("JJJJ_1")

    junc_2, cap_2 = elements[(1, 2)]
    junc_3, cap_3 = elements[(2, 3)]

    assert junc_2 == junc_3
    assert cap_2 == cap_3

    circuit = sq.Circuit(elements, flux_dist="junctions")

    circuit.description()

    assert len(circuit.parameters) == 4

    set_optim_mode(False)


def test_circuit_sampling_from_code_7() -> None:

    set_optim_mode(True)

    elements = get_elements_from_code("JJJJ_2")

    junc_1, cap_1 = elements[(0, 1)]
    junc_3, cap_3 = elements[(2, 3)]

    assert junc_1 == junc_3
    assert cap_1 == cap_3

    circuit = sq.Circuit(elements, flux_dist="junctions")

    circuit.description()

    assert len(circuit.parameters) == 4

    set_optim_mode(False)



"""Custom exceptions that can be raised during optimization. """

class ConvergenceError(Exception):
    """Error to raise when a circuit does not converge during a process, like
    optimization. Initialized with the computed value for epsilon from the
    convergence test. 
    """
    def __init__(self, epsilon):
        self.epsilon = epsilon
    def __str__(self):
        return f'Your circuit did not converge. The computed epsilon was {self.epsilon}.'



from dataclasses import dataclass
import logging
import os
from typing import Dict, Optional
import warnings

from SQcircuit import Circuit, CircuitComponent
import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch import Tensor
from tqdm import trange

from .exceptions import ConvergenceError
from .truncation import assign_trunc_nums, test_convergence
from .reparameterization import (
    get_alpha_params_from_circuit_params,
    get_circuit_params_from_alpha_params
)
from .utils import (
    init_records,
    LossFunctionType,
    RecordType,
    save_results,
    update_record,
)

logger = logging.getLogger(__name__)


def diag_with_convergence(
        circuit: Circuit,
        num_eigenvalues: int,
        total_trunc_num: int
) -> None:
    """
    Diagonalize the circuit, and, if the circuit has not converged, try
    re-allocating the truncation numbers. If this fails, then we give up and
    say the circuit has not converged; in this case a ``ConvergenceError`` is
    raised.

    Parameters
    ----------
        circuit:
            The circuit to diagonalize.
        num_eigenvalues:
            The number of eigenvalues/vectors to calculate.
        total_trunc_num:
            The maximum truncation number, to use when re-allocating.
    """

    # Reset to even split of truncation numbers
    circuit.truncate_circuit(total_trunc_num)
    circuit.diag(num_eigenvalues)
    # Check if converges with even split
    converged, _ = test_convergence(circuit, eig_vec_idx=1)
    # Otherwise try re-allocating with the heuristic function
    if not converged:
        assign_trunc_nums(circuit, total_trunc_num)
        circuit.diag(num_eigenvalues)

        converged, eps = test_convergence(circuit, eig_vec_idx=1, t=10)
        if not converged:
            raise ConvergenceError(eps)


@dataclass
class OptimizationRecord:
    """Class which contains record of optimization run.

    Sorts based on value in ``loss``.
    """
    circuit: Circuit
    loss: float
    record: RecordType

    def __lt__(self, obj):
        return self.loss < obj.loss


def run_optimization(
    circuit: Circuit,
    loss_metric_function: LossFunctionType,
    max_iter: int,
    total_trunc_num: int,
    bounds: Dict[CircuitComponent, Tensor],
    optimizer: Optimizer,
    uses_closure: bool,
    optimizer_kwargs: Optional[Dict] = None,
    scheduler : Optional[LRScheduler] = None,
    scheduler_kwargs: Optional[Dict] = None,
    num_eigenvalues: int = 10,
    save_loc: Optional[str] = None,
    identifier: Optional[str] = None,
    save_intermediate_circuits: bool = False,
) -> OptimizationRecord:
    """ Peform optimization on ``circuit`` using ``loss_metric_function`` for
    ``max_iter`` epochs, using the PyTorch ``optimizer``.

    Parameters
    ----------
        circuit:
            A circuit to optimize.
        loss_metric_function:
            Loss function to optimize.
        max_iter:
            Maximum number of iterations.
        total_trunc_num:
            Maximum total truncation number to allocate.
        bounds:
            Dictionary giving bounds for each element type.
        optimizer:
            A PyTorch ``Optimizer`` use to perform optimization.
        uses_closure:
            Whether ``optimizer`` requires a closure to step.
        optimizer_kwargs:
            An optional dictionary of keyword arguments to use when
            initializing ``optimizer``.
        scheduler:
            An optional PyTorch ``LRScheduler`` to use during optimization.
        scheduler_kwargs:
            An optional dictionary of keyword arguments to use when
            initializing ``scheduler``.
        num_eigenvalues:
            Number of eigenvalues to calculate when diagonalizing.
        save_loc:
            Optional path to folder to save results in.
        identifier:
            String identifying this run (e.g. seed, name, circuit code, …)
            to use when saving.
        save_intermediate_circuits:
            Whether to save the circuit at each iteration.
    
    Returns
    ----------
        An ``OptimizationRecord`` containing the optimized circuit, the final
        loss, and a record of the component loss and metric values at each
        epoch.
    """
    if save_loc is not None:
        if not os.path.exists(save_loc):
            raise ValueError(f'The path {save_loc} does not exist!')
        if identifier is None:
            raise ValueError('If you pass in a `save_loc` to write files to, '
                            'you must pass in an `identifier` to name the files.')
    elif save_intermediate_circuits:
        warnings.warn('Because you did not pass a `save_loc`, intermediate '
                      'circuits will not be saved.')
    if scheduler_kwargs and scheduler is None:
        warnings.warn('You did not pass in a scheduler.')


    # Set initial truncation numbers
    circuit.truncate_circuit(total_trunc_num)

    # Set up initial reparameterization
    alpha_params = get_alpha_params_from_circuit_params(
        circuit, bounds
    ).detach().clone().requires_grad_(True)

    last_total_loss = None
    last_loss_values = None
    last_metric_values = None

    optim = optimizer(
        [alpha_params],
        **optimizer_kwargs
    )

    if scheduler is not None:
        sched = scheduler(optim, **scheduler_kwargs)

    def update_circuit() -> None:
        circuit_params = get_circuit_params_from_alpha_params(
            alpha_params, circuit, bounds
        )
        circuit.parameters = circuit_params

    def objective_closure() -> Tensor:
        """Objective function for optimization with reparameterization. It 
            1. Rebuilds circuit from alpha_params
            2. Diagonalizes it and checks convergence
            3. Computes losses and metrics
        """

        # Hacky way to extract history, which is needed when using
        # optimizer with closure to avoid extra call.
        nonlocal last_total_loss
        nonlocal last_loss_values
        nonlocal last_metric_values

        # Zero the gradient
        optim.zero_grad()

        # Rebuild the circuit from the alpha parameters
        update_circuit()
        diag_with_convergence(circuit, num_eigenvalues, total_trunc_num)

        # Compute the loss…
        last_total_loss, last_loss_values, last_metric_values = loss_metric_function(circuit)
        # …and the gradient
        last_total_loss.backward()

        return last_total_loss

    # Get starting values to save
    objective_closure()

    loss_record, metric_record = init_records(
        last_loss_values,
        last_metric_values
    )

    optim_record = OptimizationRecord(circuit, last_total_loss,
                                      loss_record | metric_record)

    with trange(max_iter) as t:
        for iteration in t:
            # Update tqdm
            t.set_description(f'Iteration {iteration}')
            t.set_postfix(loss=f'{last_total_loss.detach().numpy():.3e}')
            # Log detailed information
            logger.info(
                '%s\n'
                + 'Optimization progress\n'
                + 'iteration: %s\n'
                + 'params: %s\n'
                + 'circuit params: %s\n'
                + 'loss: %s',
                90 * '-', iteration, alpha_params.detach().numpy(),
                [p.detach().numpy() for p in circuit.parameters],
                last_total_loss.detach().numpy()
            )
            for key in loss_record:
                logger.info('\t%s: %s', key, loss_record[key][-1])

            # Hold old parameter values
            old_alpha_params = alpha_params.detach().clone()

            if uses_closure:
                logger.info('Taking a step...')
                optim.step(objective_closure)
            else:
                objective_closure()
                logger.info('Taking a step...')
                logger.info('Gradient: %s.', alpha_params.grad)
                optim.step()
                update_circuit()

            if scheduler is not None:
                sched.step()

            # Update records
            update_record(loss_record, last_loss_values)
            update_record(metric_record, last_metric_values)

            optim_record.loss = last_total_loss.detach().numpy()
            optim_record.record = loss_record | metric_record

            # And save results
            if save_loc:
                save_results(
                    loss_record,
                    metric_record,
                    circuit,
                    identifier,
                    save_loc,
                    save_intermediate_circuits=save_intermediate_circuits
                )

            # Check if the optimizer has finished (params did not change)
            if torch.all(old_alpha_params == alpha_params):
                break

    return optim_record



from collections import defaultdict
from typing import Dict, List, Optional, Union

from scipy.stats import uniform, loguniform
import numpy as np
import torch

import SQcircuit as sq

from SQcircuit.circuit import Circuit
from SQcircuit.elements import (
    Capacitor,
    CircuitComponent,
    Inductor,
    Junction,
    Loop
)
from SQcircuit.settings import get_optim_mode

class CircuitSampler:
    """Class used to randomly sample different circuit configurations with a
    single inductive loop.

    Parameters
    ----------
        capacitor_range:
            A list specifying the lower bound and upper bound for the
            capacitors, in Farads.
        inductor_range:
            A list specifying the lower bound and upper bound for the inductors,
            in Henries.
        junction_range:
            A list specifying the lower bound and upper bound for the junctions,
            in Hertz.
        flux_range:
            A range of external flux values to uniformly sample from for the
            loop, in units of Phi_0.
        elems_not_to_optimize:
            An optonal list of element types **not** to optimize. Such elements
            will be randomly sampled but then fixed during the optimization
            procedure (the ``required_grad`` attribute is set to False). If
            ``None``, all element types are optimized.
    """

    def __init__(
        self,
        capacitor_range: List[float],
        inductor_range: List[float],
        junction_range: List[float],
        flux_range: List[float],
        elems_not_to_optimize: Optional[List[CircuitComponent]]=None,
    ) -> None:

        self.capacitor_range = capacitor_range
        self.inductor_range = inductor_range
        self.junction_range = junction_range
        self.flux_range = flux_range

        if elems_not_to_optimize is None:
            elems_not_to_optimize = []
        self.elems_to_optimize = [
            elem_type for elem_type in [Capacitor, Inductor, Junction, Loop]
            if elem_type not in elems_not_to_optimize
        ]

        # Special circuit codes used to create symmetric circuits
        self.special_circuit_codes = [
            'transmon',
            'flux_qubit',
            'JJJJ_1',
            'JJJJ_2'
        ]

    @property
    def bounds(self) -> Dict[CircuitComponent, Union[torch.Tensor, List[float]]]:
        """
        Bounds for circuit parameters, as a dictionary of
        ``element type: (lower bound, upper_bound)``. To match ``SQcircuit``,
        the units are
        - Capacitors: Farads
        - Inductors: Henries
        - Junctions: 2π*Hz
        - External flux: 2π*Phi_0

        Note the units of junction and external flux differ from those used to
        initialize the sampler, due to ``SQcircuit``.

        If the SQcircuit engine is ``'NumPy'``, the bounds are a list of
        floats; if it is ``'PyTorch'``, they are Tensors.
        """

        # Bounds in new units to represent difference between setting
        # and getting value in `SQcircuit`.
        flux_range_bounds = [i * 2 * np.pi for i in self.flux_range]
        junction_range_bounds = [i * 2 * np.pi for i in self.junction_range]

        if get_optim_mode():
            return {
                Capacitor: torch.tensor(self.capacitor_range, dtype=torch.float64),
                Inductor: torch.tensor(self.inductor_range, dtype=torch.float64),
                Junction: torch.tensor(junction_range_bounds, dtype=torch.float64),
                Loop: torch.tensor(flux_range_bounds, dtype=torch.float64)
            }
        else:
            return {
                Capacitor: self.capacitor_range,
                Inductor: self.inductor_range,
                Junction: junction_range_bounds,
                Loop: flux_range_bounds
            }

    def _get_elem(
        self,
        elem_str: str,
        main_loop: bool,
    ) -> Union[Capacitor, Inductor, Junction]:
        """Get a random capacitor, inductor, or junction based on the specified
        range for sampler. Element values are log-uniformly sampled.

        Parameters
        ----------
            elem_str:
                A string element type that should be either ``"L"``,``"C"``,
                or ``"J"``.
            main_loop:
                A boolean that indicates whether we are in the main loop of
                the circuit or not.

        Returns
        ----------
            An element of type specified by``elem_str``, which includes the
            loop if ``main_loop`` is ``True``.
        """

        loops = []
        if main_loop:
            loops = [self.loop,]

        if elem_str == 'J':
            if self.junction_range[0] == self.junction_range[1]:
                junc_value = self.junction_range[0]
            else:
                junc_value = loguniform.rvs(*self.junction_range)
            elem = sq.Junction(
                junc_value,
                'Hz',
                loops=loops,
                requires_grad=get_optim_mode() and Junction in self.elems_to_optimize,
            )
        elif elem_str == 'L':
            if self.inductor_range[0] == self.inductor_range[1]:
                ind_value = self.inductor_range[0]
            else:
                ind_value = loguniform.rvs(*self.inductor_range)
            elem = sq.Inductor(
                ind_value,
                'H',
                loops=loops,
                requires_grad=get_optim_mode() and Inductor in self.elems_to_optimize,
            )
        elif elem_str == 'C':
            if self.capacitor_range[0] == self.capacitor_range[1]:
                cap_value = self.capacitor_range[0]
            else:
                cap_value = loguniform.rvs(*self.capacitor_range)
            elem = Capacitor(
                cap_value,
                'F',
                requires_grad=get_optim_mode() and Capacitor in self.elems_to_optimize,
            )
        else:
            raise ValueError("elem_str should be either 'J', 'L', or 'C' ")

        return elem

    def _sample_special_circuit(self, circuit_code: str) -> Circuit:
        """Sample a symmetric circuit. Currently supported circuits are:
        - ``'transmon'``: A standard transmon (one junction and one capacitor). 
        - ``'flux_qubit'``: A qubit with three Josephson junctions where two
        of the edges (capacitor and junction) are identical.
        - ``'JJJJ_1'``: A qubit with four Josephson junctions where all
        but one edge are identical.
        - ``'JJJJ_2'``: A qubit with four Josephson junctions where opposite
        pairs of edges are identical.
        
        
        Parameters
        ----------
            circuit_code:
                A string indicating which symmetric circuit.

        Returns
        ----------
            A randomply sampled circuit with the topology defined by
            ``circuit_code``.
        """
        if circuit_code == 'transmon':
            junc_1 = self._get_elem('J', main_loop=True)
            junc_2 = junc_1

            cap = self._get_elem('C', main_loop=False)

            elements = {(0, 1): [junc_1, junc_2, cap]}

        elif circuit_code == 'flux_qubit':
            junc_1 = self._get_elem('J', main_loop=True)
            junc_2 = junc_1
            junc_3 = self._get_elem('J', main_loop=True)

            cap_1 = self._get_elem('C', main_loop=False)
            cap_2 = cap_1
            cap_3 = self._get_elem('C', main_loop=False)

            elements = {
                (0, 1): [junc_1, cap_1],
                (1, 2): [junc_2, cap_2],
                (2, 0): [junc_3, cap_3]
            }

        elif circuit_code == 'JJJJ_1':

            junc_1 = self._get_elem('J', main_loop=True)
            junc_2 = self._get_elem('J', main_loop=True)
            junc_3 = junc_2
            junc_4 = junc_3

            cap_1 = self._get_elem('C', main_loop=False)
            cap_2 = self._get_elem('C', main_loop=False)
            cap_3 = cap_2
            cap_4 = cap_3

            elements = {
                (0, 1): [junc_1, cap_1],
                (1, 2): [junc_2, cap_2],
                (2, 3): [junc_3, cap_3],
                (3, 0): [junc_4, cap_4]
            }

        elif circuit_code == 'JJJJ_2':

            junc_1 = self._get_elem('J', main_loop=True)
            junc_2 = self._get_elem('J', main_loop=True)
            junc_3 = junc_1
            junc_4 = junc_2

            cap_1 = self._get_elem('C', main_loop=False)
            cap_2 = self._get_elem('C', main_loop=False)
            cap_3 = cap_1
            cap_4 = cap_2

            elements = {
                (0, 1): [junc_1, cap_1],
                (1, 2): [junc_2, cap_2],
                (2, 3): [junc_3, cap_3],
                (3, 0): [junc_4, cap_4]
            }

        else:
            raise ValueError(f"The circuit code {circuit_code} is not "
                             "supported!")

        return sq.Circuit(elements, flux_dist='junctions')

    def _add_elem_to_elements(
        self,
        elem_str: str,
        node_1: int,
        node_2: int,
        elements: dict,
        main_loop: bool = False,
    ) -> None:
        """Add a random element to a specific edge key of elements within the
        specified range of element values.

        Parameters
        ----------
            elem_str:
                A string element type that should be either "L", "C", or "J".
            node_1:
                An integer specifying node 1 of the edge.
            node_2:
                An integer specifying node 2 of the edge.
            elements:
                A dictionary that contains the circuit's elements at each branch
                of the circuit.
            main_loop:
                A boolean that indicates whether we are in the main loop of
                the circuit or not.
        """

        elem = self._get_elem(elem_str, main_loop)

        if (node_2, node_1) in elements:
            elements[(node_2, node_1)].append(elem)
        else:
            elements[(node_1, node_2)].append(elem)

    @staticmethod
    def _if_capacitor_between_nodes(
        i: int,
        j: int,
        elements: dict
    ) -> bool:
        """Check whether there is a capacitor between nodes ``i`` and ``j``.

        Parameters
        ----------
            i:
                An integer specifying node i of the branch.
            j:
                An integer specifying node 2 of the edge.
            elements:
                A dictionary that contains the circuit's elements at each branch
                of the circuit.

        Returns
        ----------
            ``True`` if there is a capacitor between nodes ``i`` and ``j``.
        """

        if (i, j) in elements:
            if any(isinstance(elem, Capacitor) for elem in elements[(i, j)]):
                return True
        if (j, i) in elements:
            if any(isinstance(elem, Capacitor) for elem in elements[(j, i)]):
                return True

        return False

    def _add_all_to_all_cap_to_elements(
        self,
        n_nodes: int,
        elements: dict,
    ) -> None:
        """Add capacitors to all branches of the circuit.

        Parameters
        ----------
            n_nodes:
                An integer specifying the number of nodes in the circuit.
            elements:
                A dictionary that contains the circuit's elements at each branch
                of the circuit.
        """

        for i in range(n_nodes):
            for j in range(i + 1, n_nodes):
                # we prevent more than one capacitor on each branch
                if self._if_capacitor_between_nodes(i, j, elements):
                    continue

                self._add_elem_to_elements(
                    elem_str='C',
                    node_1=i,
                    node_2=j,
                    elements=elements,
                )

    @staticmethod
    def _find_last_element_index_outside_parentheses(circuit_code: str) -> int:
        """Find last element index outside parentheses. This is the last
        element that encloses the main loop.

        Parameters
        ----------
            circuit_code:
                A string specifying the circuit code.

        Returns
        ----------
            The index of the final element in the main loop.
        """
        stack = []
        last_index = -1

        for i, char in enumerate(circuit_code):
            if char == '(':
                stack.append(char)
            elif char == ')':
                if stack:
                    stack.pop()
            else:
                if not stack:
                    last_index = i

        return last_index

    def sample_circuit_code(self, circuit_code: str) -> Circuit:
        """Sample a random circuit with specified circuit code. Element
        values are log-uniformly sampled from the element ranges used to
        initialize the class.

        Parameters
        ----------
            circuit_code:
                A string specifying the circuit code.

        Returns
        ----------
            The randomly sampled circuit.
        """

        # Initialize the loop the loop
        if self.flux_range[0] == self.flux_range[1]:
            flux_value = self.flux_range[0]
        else:
            flux_value = uniform.rvs(loc=self.flux_range[0],
                                    scale=self.flux_range[1] - self.flux_range[0])
        self.loop = Loop(
            flux_value,
            requires_grad=get_optim_mode() and Loop in self.elems_to_optimize
        )

        if circuit_code in self.special_circuit_codes:
            return self._sample_special_circuit(circuit_code)

        current_node = 0
        hold_nodes = []
        idx = 0
        elements = defaultdict(list)

        final_i = self._find_last_element_index_outside_parentheses(circuit_code)

        for i in range(len(circuit_code)):
            if circuit_code[i] in ("J", "L", "C"):
                if i == final_i and i == len(circuit_code)-1:
                    self._add_elem_to_elements(
                        elem_str=circuit_code[i],
                        node_1=current_node,
                        node_2=0,
                        elements=elements,
                        main_loop=not bool(len(hold_nodes)),
                    )
                    continue
                if circuit_code[i+1] == "(":
                    if i == final_i:
                        self._add_elem_to_elements(
                            elem_str=circuit_code[i],
                            node_1=current_node,
                            node_2=0,
                            elements=elements,
                            main_loop=not bool(len(hold_nodes)),
                        )
                        hold_nodes.append(0)
                    else:
                        idx += 1
                        self._add_elem_to_elements(
                            elem_str=circuit_code[i],
                            node_1=current_node,
                            node_2=idx,
                            elements=elements,
                            main_loop=not bool(len(hold_nodes)),
                        )
                        hold_nodes.append(idx)
                elif circuit_code[i+1] == ")":
                    next_node = hold_nodes.pop()
                    self._add_elem_to_elements(
                        elem_str=circuit_code[i],
                        node_1=current_node,
                        node_2=next_node,
                        elements=elements,
                        main_loop=not bool(len(hold_nodes)),
                    )
                    current_node = next_node
                else:
                    idx += 1
                    self._add_elem_to_elements(
                        elem_str=circuit_code[i],
                        node_1=current_node,
                        node_2=idx,
                        elements=elements,
                        main_loop=not bool(len(hold_nodes)),
                    )
                    current_node = idx

        self._add_all_to_all_cap_to_elements(idx+1, elements)

        return sq.Circuit(elements, flux_dist='junctions')



"""utils.py """

import logging
import os
from typing import (
    Callable,
    Dict,
    List,
    Tuple,
    Union
)

import dill as pickle
from SQcircuit import Circuit
import torch
from torch import Tensor

logger = logging.getLogger(__name__)


################################################################################
# Helper functions.
################################################################################

def float_list(ls: List[str]) -> List[float]:
    """Evaluates a list of strings and returns a list of floats.
    """
    return [float(i) for i in ls]


# Typing
SQValType = Union[float, Tensor]
LossFunctionType = Callable[
    [Circuit],
    Tuple[Tensor, Dict[str, SQValType], Dict[str, SQValType]]
]

RecordType = Dict[
    str, Union[str, List[Tensor], List[Circuit]]
]


################################################################################
# Functions for records.
################################################################################

@torch.no_grad()
def init_records(
    loss_values: Dict[str, Tensor],
    metric_values: Dict[str, Tensor]
) -> Tuple[RecordType, RecordType]:
    """Initialize a loss and metric record, starting with a set of initial
    values.
    
    Parameters
    ----------
        loss_values:
            A dictionary of initial loss values.
        metric_values:
            A dictionary of initital metric values.

    Returns
    ----------
        loss_record:
            A dictionary of loss_name: [list of values for each epoch].
        metric_record
            A dictionary of metric_name: [list of values for each epoch].
    """
    # Init loss record
    loss_record: RecordType = {
        loss_type: [loss_value.detach().numpy()]
        for loss_type, loss_value in loss_values.items()
    }

    # Init metric record
    metric_record: RecordType = {
        metric_type: [metric_value.detach().numpy()]
        for metric_type, metric_value in metric_values.items()
    }

    return loss_record, metric_record


@torch.no_grad()
def update_record(
    record: RecordType,
    values: Dict[str, List[Tensor]]
) -> None:
    """Updates a record based on next epoch of optimization.
    
    Parameters
    ----------
        record
            A dictionary of name: [values for each epoch].
        metric_values:
            A dictionary of values for the current epoch.
    """
    for key in values.keys():
        record[key].append(values[key].detach().numpy())


def save_results(
    loss_record: RecordType,
    metric_record: RecordType,
    circuit: Circuit,
    identifier: str,
    save_loc: str,
    save_intermediate_circuits=True,
) -> None:
    """Save results from optimization to ``save_loc``.

    Parameters
    ----------
        loss_record
            The loss record from optimization.
        metric_record:
            The metric record from optimization.
        circuit:
            The current circuit.
        identifier:
            A string to use to label the saved results with.
        save_loc:
            A path to save the results at.
        save_intermediate_circuits:
            Whether to save intermediate circuits. If ``True``, the saved
            circuit record is appended to with the new ``circuit``; otherwise
            it is overwritten by ``circuit``.
    """
    save_records = {'loss': loss_record, 'metrics': metric_record}

    for record_type, record in save_records.items():
        save_url = os.path.join(
            save_loc,
            f'{record_type}_record_{identifier}.pickle'
        )
        logger.info('Saving to %s', save_url)
        with open(save_url, 'wb') as f:
            pickle.dump(record, f)

    if save_intermediate_circuits:
        write_mode = 'ab+'
    else:
        write_mode = 'wb'

    circuit_save_url = os.path.join(
        save_loc,
        f'circuit_record_{identifier}.pickle'
    )
    with open(circuit_save_url, write_mode) as f:
        pickle.dump(circuit.picklecopy(), f)



"""
Evaluate convergence test on randomly sampled (or fixed) circuit.

Usage:
  convergence_test.py <yaml_file> [--seed=<seed>] [--K=<K>]
                      [--circuit_code=<circuit_code>] [--init_circuit=<init_circuit>]
                      [--verbose]
  convergence_test.py -h | --help
  convergence_test.py --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.
  
Options:
  -h, --help    Show this screen.
  --version     Show version.

  -s, --seed=<seed>                         Seed for random generators.
  -K, --K=<K>                               Maximum truncation number.
  -c, --circuit_code=<circuit_code>         Circuit code.
  -i, --init_circuit=<init_circuit>         Set initial circuit params.
  -v, --verbose                             Turn on verbose output.
"""

import os
import random
import dill as pickle

from docopt import docopt
import matplotlib.pyplot as plt
import numpy as np
import torch

import SQcircuit as sq

import qubit_discovery as qd
from qubit_discovery import CircuitSampler
from qubit_discovery.optim.truncation import (
    assign_trunc_nums,
    get_reshaped_eigvec,
    test_convergence
)
from qubit_discovery.optim.utils import float_list
from utils import add_stdout_to_logger, load_final_circuit
from inout import add_command_line_keys, Directory, load_yaml_file

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
CONVERGENCE_REQUIRED_KEYS = [
    'seed',
    'K',
    'circuit_code',
]

CONVERGENCE_OPTIONAL_KEYS = [
    'init_circuit'
]

N_EIG_DIAG = 10
N_EIG_FLUX_SPECTRA = 10
PHI_VALUES = np.concatenate([
    np.linspace(0, 0.4, 30),
    np.linspace(0.4, 0.5, 31)[1:]
])
DEFAULT_FLUX_POINT = 0.5 - 1e-2


################################################################################
# Helper functions.
################################################################################


def set_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def calculate_flux_spectrum(circuit):
    spectrum = np.zeros((N_EIG_FLUX_SPECTRA, len(PHI_VALUES)))

    loop = circuit.loops[0]
    for flux_idx, phi in enumerate(PHI_VALUES):
        loop.set_flux(phi)
        spectrum[:, flux_idx] = circuit.diag(n_eig=N_EIG_FLUX_SPECTRA)[0].detach().numpy()

    return spectrum


def plot_flux_spectrum(spectrum, axis):
    for eigen_idx in range(N_EIG_FLUX_SPECTRA):
        axis.plot(PHI_VALUES,
                  (spectrum[eigen_idx, :] - spectrum[0, :]),
                  marker='o',
                  markersize=1.5)
    axis.axvline(x=DEFAULT_FLUX_POINT, color='b', linestyle='--')
    axis.set_xlabel(r"$\Phi_{ext}/\Phi_0$")
    axis.set_ylabel(r" $\omega_n / 2\pi$  (GHz)")


def write_test_results(axis, text):
    # bbox features
    props = dict(boxstyle='round', facecolor='yellow', alpha=0.1)
    axis.text(
        1.03, 0.98, text.strip(),
        transform=axis.transAxes,
        fontsize=12,
        verticalalignment='top',
        bbox=props
    )


def evaluate_trunc_number(circuit, trunc_nums, axis):
    if circuit.loops:
        circuit.loops[0].set_flux(DEFAULT_FLUX_POINT)
    circuit.set_trunc_nums(trunc_nums)
    circuit.diag(N_EIG_DIAG)
    passed_test, test_values = test_convergence(circuit, eig_vec_idx=1)
    spectrum = calculate_flux_spectrum(circuit)
    plot_flux_spectrum(spectrum, axis)
    output_text = f'Trunc numbers: {trunc_nums}\n'
    output_text += f'Test passed: {passed_test}\n'
    output_text += f'Test values (epsilon):\n'
    for test_value in test_values:
        output_text += f'{test_value:.3E}\n'
    write_test_results(axis, output_text)
    return passed_test


def plot_mode_magnitudes(mode_magnitudes, axis):
    mode_magnitudes = np.array(mode_magnitudes)
    axis.plot(mode_magnitudes,
              marker='o',
              linestyle='dotted',
              markersize=1.5)
    axis.set_xlabel(r"Eigenvector Entry Index $i$")
    axis.set_ylabel(r"Magnitude $|\psi_1^{[i]}|^2$")


################################################################################
# Main.
################################################################################


def main() -> None:
    ############################################################################
    # Load the Yaml file and command line parameters.
    ############################################################################

    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=CONVERGENCE_REQUIRED_KEYS,
        optional_keys=CONVERGENCE_OPTIONAL_KEYS
    )

    if arguments['--verbose']:
        add_stdout_to_logger(sq.get_logger())
        add_stdout_to_logger(qd.get_logger())

    parameters["optim_type"] = "convergence"

    directory = Directory(parameters, arguments['<yaml_file>'])
    plot_output_dir = directory.get_plots_dir()
    records_dir = directory.get_records_dir()

    ############################################################################
    # Initiate the optimization settings.
    ############################################################################

    sq.set_engine('PyTorch')

    capacitor_range = float_list(parameters['capacitor_range'])
    junction_range = float_list(parameters['junction_range'])
    inductor_range = float_list(parameters['inductor_range'])

    if "flux_range" in parameters.keys():
        flux_range = float_list(parameters['flux_range'])
        elements_not_to_optimize = []
    else:
        flux_range = [0.5, 0.5]
        elements_not_to_optimize = [sq.Loop]

    set_seed(int(parameters['seed']))

    if parameters['init_circuit'] is None:
        sampler = CircuitSampler(
            capacitor_range=capacitor_range,
            inductor_range=inductor_range,
            junction_range=junction_range,
            flux_range=flux_range,
            elems_not_to_optimize=elements_not_to_optimize
        )
        circuit = sampler.sample_circuit_code(parameters['circuit_code'])
        if circuit.loops:
            circuit.loops[0].set_flux(DEFAULT_FLUX_POINT)
        print("Circuit sampled!")
    else:
        circuit = load_final_circuit(parameters['init_circuit'])
        circuit.update()
        print("Circuit loaded!")

    fig, axes = plt.subplots(3, 3, figsize=(27, 14))
    # fig.add_gridspec(nrows=3, height_ratios=[2, 1, 1])

    ########################################################################
    # Test even distribution of truncation numbers.
    ########################################################################

    even_trunc_nums = circuit.truncate_circuit(
        int(parameters['K'])
    )
    print(f"even_trunc_nums: {even_trunc_nums}")
    even_split_passed = evaluate_trunc_number(
        circuit, even_trunc_nums, axes[0, 2]
    )

    # Plot mode magnitudes used in heuristic test (up to three modes)
    mode_magnitudes = get_reshaped_eigvec(circuit, eig_vec_idx=1)
    for mode_magnitude_idx in range(min(3, len(mode_magnitudes))):
        plot_mode_magnitudes(
            mode_magnitudes[mode_magnitude_idx],
            axes[mode_magnitude_idx, 0]
        )

    ########################################################################
    # Test heuristic truncation numbers.
    ########################################################################

    heuristic_trunc_nums = assign_trunc_nums(
        circuit,
        int(parameters['K']),
        axes=axes[:, 0],
        min_trunc_harmonic=4,
        min_trunc_charge=12,
        use_charge_heuristic=False
    )
    heuristic_passed = evaluate_trunc_number(
        circuit, heuristic_trunc_nums, axes[2, 2]
    )

    # Plot mode magnitudes used in heuristic test (up to three modes)
    mode_magnitudes = get_reshaped_eigvec(circuit, eig_vec_idx=1)
    for mode_magnitude_idx in range(min(3, len(mode_magnitudes))):
        print(
            f"mode magnitudes: {mode_magnitudes[mode_magnitude_idx].shape}"
        )

        plot_mode_magnitudes(
            mode_magnitudes[mode_magnitude_idx],
            axes[mode_magnitude_idx, 1]
        )

        ####################################################################
        # Add metadata to plot.
        ####################################################################

        props = dict(boxstyle='round', facecolor='yellow', alpha=0.1)
        # bbox features
        plt.text(1.03, 0, f"Circuit code: {parameters['circuit_code']}",
                 fontsize=12,
                 horizontalalignment='left',
                 verticalalignment='bottom',
                 transform=axes[-1, -1].transAxes,
                 bbox=props)

    ########################################################################
    # Save circuit.
    ########################################################################

    save_suffix = (
        f"{parameters['circuit_code']}"
        f"_{parameters['name']}"
        f"_{int(parameters['seed'])}"
    )
    print(f"save_suffix:{save_suffix}")
    circuit_save_url = os.path.join(
        records_dir,
        f'circuit_record_{save_suffix}.pickle'
    )
    with open(circuit_save_url, 'wb') as f:
        pickle.dump(circuit.picklecopy(), f)

    ########################################################################
    # Save test summary.
    ########################################################################

    if even_split_passed and heuristic_passed:
        test_summary = "0"
    elif not even_split_passed and heuristic_passed:
        test_summary = "1"
    elif even_split_passed and not heuristic_passed:
        test_summary = "2"
    else:
        test_summary = "3"

    summary_save_url = os.path.join(
        records_dir,
        f'summary_{save_suffix}.txt'
    )

    with open(summary_save_url, 'w') as f:
        f.write(test_summary)
    f.close()

    ########################################################################
    # Save figure.
    ########################################################################

    plt.tight_layout()
    plt.savefig(
        os.path.join(plot_output_dir, f'flux_spectra_{save_suffix}.png'),
        dpi=300,
        bbox_inches="tight"
    )


if __name__ == "__main__":
    main()



import os
import logging
import sys
from typing import Any

import dill as pickle
from matplotlib import pyplot as plt
from SQcircuit import Circuit

def add_stdout_to_logger(
        logger: logging.Logger,
        level: int =logging.INFO
) -> logging.Handler:
    """Add a handler to ``logger`` to print to stdout.
    """
    handler = logging.StreamHandler(stream=sys.stdout)
    formatter = logging.Formatter('[%(asctime)s] %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')

    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)

    return handler


def load_record(url: str) -> Any:
    if not os.path.exists(url):
        print(f"The file '{url}' does not exist.")
        return None

    try:
        with open(url, 'rb') as f:
            record = pickle.load(f)
        return record

    except EOFError:
        print(f"Warning: The file '{url}' is empty or corrupted.")
        return None


def load_initial_circuit(circuit_record: str) -> Circuit:
    with open(circuit_record, 'rb') as f:
        first_circ = pickle.load(f)
    return first_circ


def load_final_circuit(circuit_record: str) -> Circuit:
    with open(circuit_record, 'rb') as f:
        try:
            while True:
                last_circ = pickle.load(f)
        except EOFError:
            pass
    return last_circ


def load_all_circuits(circuit_record: str) -> Circuit:
    circuits = []
    with open(circuit_record, 'rb') as f:
        try:
            while True:
                next_circuit = pickle.load(f)
                circuits += [next_circuit, ]
        except EOFError:
            pass
    return circuits


def set_plotting_defaults(single_color=False):
    plt.rcParams['lines.linewidth'] = 2.2
    plt.rcParams["xtick.major.size"] = 5
    plt.rcParams["xtick.major.width"] = 1.2
    plt.rcParams["ytick.major.size"] = 5
    plt.rcParams["ytick.major.width"] = 1.2
    plt.rcParams["axes.titlesize"] = 24
    plt.rcParams["font.size"] = 24
    plt.rcParams['xtick.labelsize'] = 15
    plt.rcParams['ytick.labelsize'] = 15
    color_list = ['#8D1514', '#007662', '#333131', '#D1C295', '#FD7C34', '#6B3B6D']
    if single_color:
        color_list = ['#8D1514', ]
    plt.rcParams['axes.prop_cycle'] = plt.cycler(color=color_list)



import SQcircuit as sq

import qubit_discovery as qd
from qubit_discovery.optim.truncation import (
    assign_trunc_nums,
    test_convergence as convergence_test # rename so pytest ignores
)

N_EIG = 10

def test_JLL():
    loop = sq.Loop(0.35)
    circuit = sq.Circuit(
        {
            (0, 1): [sq.Junction(61, 'GHz', loops=[loop]), sq.Capacitor(4e-13, 'F')],
            (1, 2): [sq.Inductor(5e-9, 'H', loops=[loop]), sq.Capacitor(4e-13, 'F')],
            (2, 0): [sq.Inductor(1e-9, 'H', loops=[loop]), sq.Capacitor(1e-13, 'F')],
        },
        flux_dist='junctions'
    )

    K = 700
    circuit.truncate_circuit(K)
    circuit.diag(N_EIG)

    assert circuit.trunc_nums == [26, 26]
    assert not circuit.check_convergence(t=5)[0], 'Circuit should not converge'

    qd.optim.assign_trunc_nums(circuit, total_trunc_num=K)
    circuit.diag(N_EIG)

    assert circuit.trunc_nums == [9, 77]
    assert circuit.check_convergence(t=5)[0], 'Circuit should converge'


def test_JJJ_even():
    """Test the default behavior in the case of charge modes (which is to 
    just assign truncation numbers evenly).
    """
    J = sq.Junction(5, 'GHz')
    C = sq.Capacitor(1, 'GHz')
    circuit = sq.Circuit(
        {
            (0, 1): [J, C],
            (1, 2): [J, C],
            (2, 0): [J, C],
        }
    )

    K = 1000
    circuit.truncate_circuit(K)
    assert circuit.trunc_nums == [16, 16]

    circuit.diag(N_EIG)

    qd.optim.assign_trunc_nums(circuit, total_trunc_num=K)

    assert circuit.trunc_nums == [16, 16]


def build_paper_circuit():
    loop = sq.Loop(0.5)

    JJ_10 = sq.Junction(10.83, 'GHz', loops=[loop])
    C_10 = sq.Capacitor(7.974e-15, 'F')
    L_12 = sq.Inductor(1.53e-06, 'H', loops=[loop])
    C_12 = sq.Capacitor(3.989e-14, 'F')
    JJ_23 = sq.Junction(33.96, 'GHz', loops=[loop])
    C_23 = sq.Capacitor(4.79e-14, 'F')
    L_03 = sq.Inductor(5.482e-13, 'H', loops=[loop])
    C_03 = sq.Capacitor(4.806e-15, 'F')
    C_02 = sq.Capacitor(6.298e-13, 'F')
    C_13 = sq.Capacitor(3.32e-13, 'F')

    elements = {
        (0, 1): [JJ_10, C_10],
        (1, 2): [L_12, C_12],
        (2, 3): [JJ_23, C_23],
        (0, 3): [L_03, C_03],
        (0, 2): [C_02],
        (1, 3): [C_13]
    }

    return sq.Circuit(elements)

def test_paper_circuit() -> None:
    circuit = build_paper_circuit()
    K = 6000
    circuit.truncate_circuit(K)
    print('Even truncation', circuit.m)
    assert circuit.m == [18, 18, 17]

    circuit.diag(N_EIG)
    assign_trunc_nums(
        circuit,
        K,
        min_trunc_harmonic=4
    )
    print('Heuristic truncation', circuit.m)
    assert circuit.m == [4, 88, 17]

    # Need to set back the truncation numbers to match what we had
    # during diagonalization.
    circuit.truncate_circuit(K)

    assign_trunc_nums(
        circuit,
        K,
        min_trunc_harmonic=4,
        min_trunc_charge=12
    )
    print('Heuristic truncation with charge mininum', circuit.m)
    assert circuit.m == [4, 65, 23]

    circuit.diag(N_EIG)
    passed_test, test_values = convergence_test(circuit, eig_vec_idx=1)
    print('Convergence test values', test_values)
    assert passed_test, "Test should pass"



from .BFGS import run_BFGS
from .SGD import run_SGD
from .exceptions import *
from .optim import run_optimization
from .sampler import CircuitSampler
from .truncation import (
    assign_trunc_nums,
    charge_mode_heuristic,
    harmonic_mode_heuristic,
    trunc_num_heuristic
)



"""
Optimize a single circuit.

Usage:
  optimize.py <yaml_file> [--seed=<seed>] [--circuit_code=<circuit_code>]
              [--optim_type=<optim_type>] [--init_circuit=<init_circuit>]
              [--save-intermediate] [--verbose]
  optimize.py -h | --help
  optimize.py --version

Arguments:
  <yaml_file>   YAML file containing details about the optimization.

Options:
  -h, --help    Show this screen.
  --version     Show version.

  -c, --circuit_code=<circuit_code>         Code for circuit topology.
  -s, --seed=<seed>                         Seed (integer) for random generators.
  -o, --optim_type=<optim_type>             Optimization method to use.
  -i, --init_circuit=<init_circuit>         Set intial circuit to <init_circuit>.
  --save-intermediate                       Save intermediate circuits during
                                            optimization to file.
  -v, --verbose                             Turn on verbose output.

Notes: Optional arguments to optimize.py must either be provided on the
command line or in <yaml_file>.
"""

import random

from docopt import docopt
import torch
import numpy as np

import qubit_discovery as qd
from qubit_discovery import (
    build_loss_function,
    CircuitSampler,
    run_BFGS,
    run_SGD
)
from qubit_discovery.optim.utils import float_list
import SQcircuit as sq

from utils import add_stdout_to_logger, load_final_circuit
from inout import add_command_line_keys, Directory, load_yaml_file

################################################################################
# General Settings.
################################################################################

# Keys that should be in either command line or Yaml file.
OPTIMIZE_REQUIRED_KEYS = [
    'seed',
    'circuit_code',
    'optim_type',
    'save-intermediate',
]

OPTIMIZE_OPTIONAL_KEYS = [
    'init_circuit'
]

################################################################################
# Helper functions.
################################################################################


def set_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

################################################################################
# Main.
################################################################################


def main() -> None:

    ############################################################################
    # Loading the Yaml file and command line parameters.
    ############################################################################

    arguments = docopt(__doc__, version='Optimize 0.8')

    parameters = load_yaml_file(arguments['<yaml_file>'])

    parameters = add_command_line_keys(
        parameters=parameters,
        arguments=arguments,
        keys=OPTIMIZE_REQUIRED_KEYS,
        optional_keys=OPTIMIZE_OPTIONAL_KEYS
    )

    directory = Directory(parameters, arguments['<yaml_file>'])

    if arguments['--verbose']:
        add_stdout_to_logger(sq.get_logger())
        add_stdout_to_logger(qd.get_logger())

    ############################################################################
    # Initiating the optimization settings.
    ############################################################################

    sq.set_engine('PyTorch')
    sq.set_max_eigenvector_grad(2)

    capacitor_range = float_list(parameters['capacitor_range'])
    junction_range = float_list(parameters['junction_range'])
    inductor_range = float_list(parameters['inductor_range'])

    if 'flux_range' in parameters.keys():
        flux_range = float_list(parameters['flux_range'])
        elements_not_to_optimize = []
    else:
        flux_range = [0.5, 0.5]
        elements_not_to_optimize = [sq.Loop]

    sampler = CircuitSampler(
            capacitor_range=capacitor_range,
            inductor_range=inductor_range,
            junction_range=junction_range,
            flux_range=flux_range,
            elems_not_to_optimize=elements_not_to_optimize
    )

    set_seed(int(parameters['seed']))

    my_loss_function = build_loss_function(
        use_losses=parameters['use_losses'],
        use_metrics=parameters['use_metrics']
    )

    if parameters['init_circuit'] is None or parameters['init_circuit'] == '':
        circuit = sampler.sample_circuit_code(parameters['circuit_code'])
        print('Circuit sampled!')
    else:
        circuit = load_final_circuit(parameters['init_circuit'])
        circuit.update()
        print('Circuit loaded!')

    circuit.truncate_circuit(parameters['K'])

    ############################################################################
    # Run the optimizations.
    ############################################################################

    if parameters['optim_type'] == 'SGD':
        run_SGD(
            circuit=circuit,
            loss_metric_function=my_loss_function,
            max_iter=parameters['epochs'],
            total_trunc_num=parameters['K'],
            bounds=sampler.bounds,
            num_eigenvalues=parameters['num_eigenvalues'],
            identifier = f'{parameters["circuit_code"]}_{parameters["name"]}_SGD_{parameters["seed"]}',
            save_loc=directory.get_records_dir(),
            save_intermediate_circuits=parameters['save-intermediate']
        )
    elif parameters['optim_type'] == 'BFGS':
        run_BFGS(
            circuit=circuit,
            loss_metric_function=my_loss_function,
            max_iter=parameters['epochs'],
            total_trunc_num=parameters['K'],
            bounds=sampler.bounds,
            num_eigenvalues=parameters['num_eigenvalues'],
            identifier = f'{parameters["circuit_code"]}_{parameters["name"]}_BFGS_{parameters["seed"]}',
            save_loc=directory.get_records_dir(),
            save_intermediate_circuits=parameters['save-intermediate']
        )
    else:
        raise ValueError(f'Optimization with {parameters["optim_type"]} '
                         'is not supported.')


if __name__ == '__main__':
    main()
